import sys, os
from math import log, exp, log10
from datetime import date
from time import clock, localtime
import random
from copy import deepcopy
import utilities


class SingleDescriptor:
	def __init__(self, name, weight=0., norm=2., normmethod=None, distance='euclidean', limit=False, simultaneous=None, energyWeight=False, origin='DESCRIPTORS', neededBy=['target', 'corpus'], packagename=None):
		self.singleNumberDescriptors = ['effDur-seg', 'peakTime-seg', 'MIDIPitch-seg', 'percentInFile-seg', 'temporalIncrease-seg', 'temporalDecrease-seg', 'logAttackTime-seg', 'temporalCentroid-seg']
		self.neverRecalculate = ['zeroCross', 'f0', 'peakamp', 'peakfrq']

		self.name = name
		self.weight = weight
		self.norm = norm
		self.normmethod = normmethod
		self.distance = distance
		self.limit = limit
		self.neededBy = neededBy
		self.origin = origin
		self.energyWeight = energyWeight
		self.tgt_modify = []
		self.packagename = packagename

		if name.find('-slope-seg') != -1:
			self.type = 'slope-regression'
			self.seg = True
		elif name.find('-mean-seg') != -1:
			self.type = 'mean'
			self.seg = True
		elif name.find('-seg') != -1:
			self.type = 'segmented'
			self.seg = True
		elif name.find('-odf-') != -1:
			self.type = 'onsetdetection'
			self.seg = False
		elif name.find('-delta') != -1:
			self.type = 'delta'
			self.seg = False
		elif name.find('-deltadelta') != -1:
			self.type = 'deltadelta'
			self.seg = False
		else:
			self.type = 'rawsdif'
			self.seg = False

		if self.seg:
			if self.type == 'slope-regression': self.seg_method = 'slope'
			elif self.type == 'mean': self.seg_method = 'mean'
			elif self.name in ["power-seg", "rms-seg"]: self.seg_method = 'max'
			elif self.name in self.singleNumberDescriptors: self.seg_method = 'single_number'
			else: self.seg_method = 'weighted_mean'

		# get the "parent descriptors" which this descriptor depend upon
		namesplit = self.name.split('-')
		self.parents = []
		if self.name in self.singleNumberDescriptors: pass
		elif self.type in ['slope-regression', 'onsetdetection', 'delta', 'deltadelta', 'mean']:
			self.parents = [namesplit[0]]
		elif self.type == 'segmented':
			self.parents = []
			cnt = 1
			while True:
				if cnt > len(namesplit)-1: break
				self.parents.append( '-'.join(namesplit[:cnt]) )
				cnt += 1
		#print self.name, self.parents, self.singleNumberDescriptors
		self.describes_energy = False
		self.is_mixable = False
		if SdifInterface.agDescriptToSdif.has_key(name):
			filetype, energyBool, mixBool, frame, matrix, row, col = SdifInterface.agDescriptToSdif[name]
			self.describes_energy = energyBool
			self.is_mixable = mixBool
		else:
			for pname in self.parents:
				filetype, energyBool, mixBool, frame, matrix, row, col = SdifInterface.agDescriptToSdif[pname]
				if energyBool: self.describes_energy = True
				if not mixBool: self.is_mixable = False
		#print name, self.describes_energy, self.is_mixable
#		sys.exit()
	###########
	def __repr__(self):
		return self.name
	###########
	def get_norm_coeffs(self, tgt, cpshandles, datasource):
		tmpcps = []
		if self.seg:
			tmptgt = filter(None, tgt.data[datasource][self.name])
			for h in cpshandles: tmpcps.append( h.data[datasource][self.name][0] )
		else:
			tmptgt = tgt.data[datasource][self.name]
			for h in cpshandles: tmpcps.append( h.data[datasource][self.name] )
			tmpcps = np.concatenate(tmpcps)
		
		if self.norm == 1: # togerther norm
			both = np.concatenate((tmptgt, tmpcps))
			if self.normmethod == 'stddev':
				self.tgt_coeff = np.mean( both ), np.std( both )
			elif self.normmethod == 'minmax':
				minny = np.min( both )
				range = np.max( both )-minny
				self.tgt_coeff = minny, range
			self.cps_coeff = self.tgt_coeff[:]
		elif self.norm == 2: # separate norm
			if self.normmethod == 'stddev':
				self.tgt_coeff = np.mean( tmptgt ), np.std( tmptgt )
				self.cps_coeff = np.mean( tmpcps ), np.std( tmpcps )
			elif self.normmethod == 'minmax':
				minny = np.min( tmptgt )
				self.tgt_coeff = minny, np.max(tmptgt)-minny
				minny = np.min( tmpcps )
				self.cps_coeff = minny, np.max(tmpcps)-minny
		#print self.name, self.norm,  self.normmethod, self.tgt_coeff, self.cps_coeff
		del tmptgt
		del tmpcps





		
class DescriptorManager:
	def __init__(self):
		# special names that indicate combinations of desciptors
		self.packagedDesciptors = ['mfccs', 'chromas', 'peakfrqs', 'loudestpeak', 'crests', 'flatnesses', 'harmonictristimuluses', 'perceptualtristimuluses', 'cluster_cents', 'autocoors']
		# don't ever weight by energy
		self.describesAmplitude = ['power']  
		# these descriptors can't be mixed, seg methods included
		self.neverMix = ['effDur-seg', 'zeroCross', 'f0', 'dur-seg', 'MIDIPitch-seg', 'percentInFile-seg', 'peakamp', 'peakfrq', 'logAttackTime-seg'] # matches partial strings!
		#
		self.singleNumberDescriptors = ['effDur-seg', 'peakTime-seg', 'MIDIPitch-seg', 'percentInFile-seg', 'temporalIncrease-seg', 'temporalDecrease-seg', 'logAttackTime-seg', 'temporalCentroid-seg']
		# here are things not attempted to evalute by SdifInterfaceescripor
		self.descriptorsLoadIgnore = ['loudestpeak']
		# here get added metrics needed for analysis and subtraction but not called for by the user
		self.internalMetrics = ['power', 'peakTime-seg', 'power-seg', 'power-mean-seg', 'effDur-seg', 'f0-seg']
		# here get added metrics needed for the corpus
		self.internalSrcMetrics = ['MIDIPitch-seg']  # used for musical writing/score outputs
		# check to see if midipitch method requires a descriptor
		for c in ops.CORPUS:
			if type(c) == dict and c.has_key('MIDIPitch') and c['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and c['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(c['MIDIPitch'])
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('MIDIPitch') and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'])
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('MIDIPitch') and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] in ['centroid-seg', 'f0-seg'] and ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'] not in self.internalMetrics:
				self.internalMetrics.append(ops.CORPUS_GLOBAL_ATTRIBUTES['MIDIPitch'])
		# transpose
		self.internalMetrics = set(self.internalMetrics) # remove and repetitions
		
		self.limitsFromCorpus = {}
		# add any global limits to limitList
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('limit'):
			for stry in ops.CORPUS_GLOBAL_ATTRIBUTES['limit']:
				dname,symb,val = utilities.parseEquationString(stry, ['==', '!=', '<', '<=', '>', '>='])
				self.limitsFromCorpus[dname] = True
		# loop through corpus items and add limits to the limit list
		for idx, entry in enumerate(ops.CORPUS):
			if type(entry.limit) == str: entry.limit = [entry.limit] # support strings not in lists
			for stry in entry.limit:
				dname,symb,val = utilities.parseEquationString(stry, ['==', '!=', '<', '<=', '>', '>='])
				self.limitsFromCorpus[dname] = True
		##########################
		## NEW FILL DESCRIPTORS ##
		##########################
		self.master_list = []
		self.master_name_list = []
		##############################################
		## make desciptor similarity matching steps ##
		##############################################		
		for sidx, spassobj in enumerate(ops.SEARCH):
		#	def __init__(self, descriptor_dict, method='closest', percent=None):
			for dobj in spassobj.descriptor_list:
				added = self.addDescriptor(dobj.name, weight=dobj.weight, norm=dobj.norm, normmethod=dobj.normmethod, distance=dobj.distance, energyWeight=dobj.energyWeight, origin='SEARCH_PASS%i'%sidx, neededBy=['target', 'corpus'])
				for d in added:
					if d.seg: spassobj.seg.append(d)
					else: spassobj.timevary.append(d)


		for origin, dicty in [('TARGET_ONSET_DESCRIPTORS', ops.TARGET_ONSET_DESCRIPTORS)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=dicty[desc], norm=None, distance=None, simultaneous='subtract', origin=origin, neededBy=['target'])

		for origin, dicty in [('LOAD_ADDITIONAL_DESCRIPTORS', ops.LOAD_ADDITIONAL_DESCRIPTORS), ('AUDIOGUIDE_INTERNAL', self.internalMetrics)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=0, norm=None, distance=None, origin=origin, neededBy=['target', 'corpus'])

		for origin, dicty in [('AUDIOGUIDE_INTERNAL_CORPUS', self.internalSrcMetrics), ('ORDER_CORPUS_BY_DESCRIPTOR', ops.ORDER_CORPUS_BY_DESCRIPTOR), ('CORPUS_RESTRICT_DISCRIPTOR_IN_TIME', ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME)]:
			for desc in dicty:
				self.addDescriptor(desc, weight=0, norm=None, distance=None, origin=origin, neededBy=['corpus'])

		for origin, listy in [('limitsFromCorpusOps', self.limitsFromCorpus)]:
			for desc in listy:
				self.addDescriptor(desc, weight=0, norm=None, distance=None, limit=True, origin=origin, neededBy=['corpus'])

		# done with lists
		self.getlists()
		#self.getsimcalclist()
	###########
	def addDescriptor(self, name, weight=0., norm=None, normmethod=None, distance=None, limit=False, simultaneous=None, origin=None, neededBy=None, energyWeight=False):
		packagename, descriptorList = metrics.name2packageExpansion(name, weight, neededBy)
		tmp = []
		for dname, weight, neededBy in descriptorList:
			dobj = SingleDescriptor(dname, weight=weight, norm=norm, normmethod=normmethod, limit=limit, distance=distance, origin=origin, energyWeight=energyWeight, neededBy=neededBy, packagename=packagename)
			self.master_list.append( dobj )
			self.master_name_list.append(dname)
			tmp.append(dobj)
		return tmp
	###########
	def addParents(self):		
		for d in self.master_list:
			for pname in d.parents:
				if pname in self.master_name_list: continue
				self.addDescriptor(pname, weight=0., norm=None, distance=None, origin='PARENT', neededBy=d.neededBy)
	###########
	def getDescriptorObjByName(self, name):
		return self.master_list[self.master_name_list.index(name)]
	###########
	def getlists(self):
		self.target_load = []
		self.target_seg = []
		self.target_onfly = []
		self.target_descriptor_subtract = []
		self.target_descriptor_mixture = []
		#self.corpus_load = []
		#self.corpus_limit = []
		#self.corpus_onfly = []
		#self.normalize_compute = []
		
		self.addParents()
		# do mixture list!
		for d in self.master_list:
			tmp_power_list = [] # because power descriptors must come last in the procession
			if d.origin.find('SEARCH_PASS') == -1: continue
			mix_me = True
			if d.weight == 0: mix_me = False
			for skip in self.neverMix:
				if d.name.find(skip) != -1: mix_me = False
			if not mix_me: continue
			descripts = []
			for pname in d.parents: descripts.append(self.getDescriptorObjByName(pname))
			descripts.append(d) # goes after parent descriptors
			# make sure power if last!
			if d.name in self.describesAmplitude: tmp_power_list.extend(descripts)
			else: self.target_descriptor_mixture.extend(descripts)
		if len(tmp_power_list) == 0:
			# we MUST have power in there, even if it is not in the distance metrics
			self.target_descriptor_mixture.append(self.getDescriptorObjByName('power'))
		else:
			self.target_descriptor_mixture.extend(tmp_power_list)
		
		# do subtraction list!
		for d in self.master_list:
			if d.origin not in ['TARGET_ONSET_DESCRIPTORS']: continue
			for string in self.describesAmplitude:
				if d.name.find(string) == -1: continue
				for pname in d.parents:					
					self.target_descriptor_subtract.append(self.getDescriptorObjByName(pname))
				self.target_descriptor_subtract.append(d)
		
		for d in self.master_list:
			if 'target' in d.neededBy:
				if d.type == 'rawsdif': self.target_load.append(d)
				elif d.type in ['segmented', 'slope-regression']: self.target_seg.append(d)
				else: self.target_onfly.append(d)
			#if 'corpus' in d.neededBy:
			#	if d.type == 'rawsdif': self.corpus_load.append(d)
			#	else: self.corpus_onfly.append(d)
			#if d.norm != None:
			#	self.normalize_compute.append(d)
			#if d.limit:
			#	self.corpus_limit.append(d)
		# do target feature modification
		for mstr in ops.TARGET_DESCRIPTOR_MODIFY:
			if -1 != mstr.find('-seg') or -1 != mstr.find('-slope'):
				print "\n\nERROR, Can only use time-varying descriptors for target descriptor modification.\n\n"
				sys.exit()
			for mathsymbol in ['-', '+', '/', '*', '**']:
				if mstr.find(mathsymbol) != -1:
					pieces = mstr.split(mathsymbol)
					dname = pieces[0].strip()
					value = float(pieces[-1])
					break
			d = self.getDescriptorObjByName(dname)
			d.tgt_modify.append([mathsymbol, value])
		# done
		self.target_load = set(self.target_load)
		self.target_seg = set(self.target_seg)
		self.target_onfly = set(self.target_onfly)
		#self.corpus_load = set(self.corpus_load)
		#self.corpus_onfly = set(self.corpus_onfly)
		#self.normalize_compute = set(self.normalize_compute)
		p.write("DESCRIPTORS: load from sdif %s\n"%str(self.target_load))
	############################################################################
	def allDescriptorsList(self, SdifInterface):
		self.allSegmentMethods = []
		for name in SdifInterface.agDescriptToSdif.keys():
			add_me = True
			for string in self.singleNumberDescriptors:
				if string.find(name) != -1: add_me = False
			if add_me: self.allSegmentMethods.append(name+'-seg')
		# write me if I don't exist!
		textPath = os.path.join(os.path.dirname(__file__), 'all-descriptors.txt')
		if os.path.exists(textPath): return
		fh = open(textPath, 'w')
		all_d = []
		single_n_d = self.singleNumberDescriptors[:]
		max_len = 0
		for d in SdifInterface.agDescriptToSdif.keys():
			if d.find('-seg') != -1:
				single_n_d.append(d)
				continue
			if len(d) > max_len: max_len = len(d)
			all_d.append(d)
		all_d.sort()
		col_size = max_len+6
		fh.write('--------------------------------------------------------------\n')
		fh.write('----------------------  SINGLE NUMBER  -----------------------\n')
		fh.write('These descriptors consist of a single number for a segment of sound.  they are useful in temporal modeling -- i.e., effDur-seg describes the effectuve duration of a trget or corpus segment.  note that "effDur" does not exist -- this is why it is called a single number descriptor\n\n')
		for d in single_n_d:
			fh.write('%s\n'%d)
		fh.write('\n\n')
		fh.write('--------------------------------------------------------------\n')
		fh.write('-------------------  PACKAGED DESCRIPTOR  --------------------\n')
		fh.write('These descriptors refer to a group of other descritors with (sometimes) differing weights.  They are provided for convenience.  Here, the package descriptor name is shown to equal a list of "child" descriptors plus their weights.\n\n')
		for d in self.packagedDesciptors:
			name, listy = metrics.name2packageExpansion(d, 1, ['target'])
			pack = []
			for l in listy:
				pack.append('(%s %2.2f)'%(l[0], l[1]))
			fh.write('%s = %s\n\n'%(d, ', '.join(pack)))
		fh.write('\n\n')
		fh.write('--------------------------------------------------------------\n')
		fh.write('-----------------------  TIME VARYING  ------------------------\n')
		fh.write('These descriptors vary in time since each analysis frame produces a different value.  Note that any descriptor here, for example "centroid", may have different ending appended to it alter its values.\n\n')
		for d in all_d:
			fh.write('%s -- also: %s-delta, %s-deltadelta, %s-seg, %s-slope-seg\n'%(d, d, d, d, d))
		fh.close()







################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class analData: # holds data and stuff for database members
	def __init__(self):
		self.pb = None # init pb printing
		#self.setSampleRate()
		self.scoresToRender = []
		self.allHandles = [] # a list of all sample handles
		self.sourceHandles = [] # a list of all database sample handles
		self.hannWindow = metrics.hannWin(2096)
		self.negEnergy = 0
		self.analNormalizeDict = {}
		self.srcMaxPowers = []
		self.corpusLengthInSeconds = 0
		self.normPrintData = {}
		self.selectedSegmentTimes = []
		self.selectedSegments = [] # maintain a list of selected corpus segments
		self.activeSegments = [] # maintain a list of currently active corpus segments
		self.countRanOutOfSamples = 0
		self.logicForSegmentTermination = {'max overlapping segments': 0, 'max segments per onset': 0, 'below onset threshold': 0, 'out of valid samples': 0, 'SEARCH conditions not met': 0,}
		self.timerDict = {}
		self.limitAnalStats = {}
		self.limitObjList = []
	############################################################################
	def f2s(self, frame):
		'''descriptor frame to time in seconds'''
		sec = (frame*ops.DESCRIPTOR_HOP_SIZE_SEC) + (ops.DESCRIPTOR_WIN_SIZE_SEC / 2.0)
		sec = (frame*ops.DESCRIPTOR_HOP_SIZE_SEC)
		#print "\nFRAME %i to SECOND %f\n\n"%(frame, sec)
		return sec
	############################################################################
	def s2f(self, sec):
		fr = int((sec-(ops.DESCRIPTOR_WIN_SIZE_SEC/2.0))/ops.DESCRIPTOR_HOP_SIZE_SEC)
		fr = int(sec/ops.DESCRIPTOR_HOP_SIZE_SEC)
		'''time in seconds to descriptor frame'''
		#print "\nSECOND %f to FRAME %i\n\n"%(sec, fr)
		return fr
	############################################################################
	def timer(self, name, call):
		from time import clock
		if call == "set":
			self.timerDict[name] = clock()
			return
		elif call == "get":
			self.timerDict[name] = clock()-self.timerDict[name]
			return self.timerDict[name]
	############################################################################
	def printTimeData(self):
		postLevel = 2
		totalTime = 0
		for key, val in self.timerDict.iteritems():
			totalTime += val
		p.middle('Elapsed Time - Total '+utilities.trunc(totalTime, 2)+" seconds", postLevel)
		for key, val in self.timerDict.iteritems():
			p.post(str(key)+" -> ${RED}"+utilities.trunc(val, 2)+"${NORMAL} seconds", postLevel)
		p.post("", postLevel)
	############################################################################
#	def segTimesToWindowedLength(self, h):
#		if h.startSec == None: h.startSec = 0.
#		if h.endSec == None:
#			h.endSec = h.sfdur
#		print h.startSec, h.endSec, h.filename
#		print SdifInterface.sec2frametimes(h.startSec, h.endSec, h.filename)
#		
#		h.lengthInFrames = SdifInterface.sec2frame(h.endSec-h.startSec)
#
#		if h.lengthInFrames < 1: # at least one frame!
#			h.endSec = h.startSec+ops.DESCRIPTOR_HOP_SIZE_SEC
#			h.lengthInFrames = 1
	############################################################################
#	def segmentTimes(self, handle, sfdur):
#		if handle.startSec in [None,"start","Start"]: handle.startSec = 0 # if no startSec set to 0
#		if handle.endSec in [None,"end","End"]: handle.endSec = sfdur
#		if handle.endSec > sfdur: handle.endSec = sfdur
	############################################################################
	def makeTargetAttenuationGrid(self, handle):
		#############################
		## TARGET GRID ATTENUATION ##
		#############################
		# test to see if we need to make a temporal grid overlayed on the target's amplitude
		if type(ops.TARGET_MAKE_GRID_ATTENUATION) == dict and handle.typed == 'target':
			# make articulation points with randomness if specified
			attackPointsFrames = []
			peakAmps = []
			t = 0
			while t+ops.TARGET_MAKE_GRID_ATTENUATION['duration'] < ag.f2s(tgt.lengthInFrames):
				rndAdd = random.uniform(ops.TARGET_MAKE_GRID_ATTENUATION['randomnessSec']/-2., ops.TARGET_MAKE_GRID_ATTENUATION['randomnessSec']/2.)
				time = round(ag.s2f(t+rndAdd), 0)
				attackPointsFrames.append(time)
				peakAmps.append(tgt.data['raw']['power'][time])
				t += ops.TARGET_MAKE_GRID_ATTENUATION['period']
			peakMaxAmp = np.max(peakAmps)
			# make the maksing array for tgt power multiplication
			maskArray = np.zeros(tgt.lengthInFrames)
			for onset in attackPointsFrames:
				onset = int(onset)
				offset = ag.s2f(onset+(ops.TARGET_MAKE_GRID_ATTENUATION['duration']))
				singleGrid = np.linspace(1., 0., num=offset-onset)
				# add exponential slopping
				singleGrid = np.power(singleGrid, ops.TARGET_MAKE_GRID_ATTENUATION['slope'])
#				print onset, offset
				maskArray[onset:offset] = singleGrid
				# if normalize, make all grid points have then same peak amp
				if ops.TARGET_MAKE_GRID_ATTENUATION['norm']:
					boost = peakMaxAmp/tgt.data['raw']['power'][onset]
					maskArray[onset:offset] *= boost
			# mask it
			tgt.data['raw']['power'] *= maskArray
	############################################################################
	def calculatePostSelectionError(self, s, e, selectedSegments):
		return 0.
		#for d in descriptorData.rvw
		error = 0.
		return error
	############################################################################
#	def segAnal(self, handle, d, startFrame, endFrame):
		# a dictionary with (start,end) tuple as key and descriptor as value
#		handle.data['raw'][d.name][startFrame] = metrics.analSegment(d, handle, startFrame, endFrame, "Seg", ag)[0]
	############################################################################
	def clearSegDict(self, pointerToDict, time, dur):
		for (st,ed) in pointerToDict.keys():
			if st >= time and ed <= time+dur:
				del pointerToDict[(st,ed)]
	############################################################################
	def updateActiveSegments(self, currentTime, length):
		activeSegments = []
		for segment in self.selectedSegments:
			if segment[0] < currentTime+length and segment[1] > currentTime: # include it!
				activeSegments.append(segment)
		return activeSegments	
	############################################################################
#	def postNormalizePrint(self, norm_compute):
#		printLevel = 2
#		p.middle('Normalization Data', 2)
#		for d in norm_compute:#, data in self.normPrintData.iteritems():
#			data = self.normPrintData[d.name]
#			printString = ""
#			printString += "${RED}"+d.name.upper()+"${NORMAL}: "
#			if data['sign'] == '-': printString += "${GREEN}INVERTED${NORMAL} and "
#			printString += "normalized "
#			if data['percent'] != 100 and data['norm'] == 'between': printString += str(data['percent'])+"% "
#			if data['norm'] == 'together':
#				printString += "${YELLOW}COMBINDING${NORMAL} the TARGET and SRC data "
#			elif data['norm'] == 'separate':
#				printString += "${YELLOW}SEPARATING${NORMAL} the TARGET and SRC data "
#			elif data['norm'] == 'between':
#				printString += "${YELLOW}INTERPOLATING BETWEEN${NORMAL} the TARGET and SRC data "
#
#			if data['percent'] != 100 and data['norm'] != 'between':
#				printString += "${GREEN}"
#				if data['percent'] > 100:
#					printString += "; expanding at "
#				elif data['percent'] < 100:
#					printString += "; compressing at "
#				printString += str(data['percent'])+"% "
#				printString += "${NORMAL}"
#
#			p.post(printString, printLevel)
#			if self.targetNormCoeffs[d.name] != self.sourceNormCoeffs[d.name]:
#				p.post("\tTARGET "+d.name+": mean->"+utilities.trunc(self.targetNormCoeffs[d.name][0], 3)+" std->"+utilities.trunc(self.targetNormCoeffs[d.name][1], 3)+"\tDATABASE "+d.name+": mean->"+utilities.trunc(self.sourceNormCoeffs[d.name][0], 3)+" std->"+utilities.trunc(self.sourceNormCoeffs[d.name][1], 3)+"", printLevel)
#			else:
#				p.post("\tTARGET+DATABASE "+d.name+": mean->"+utilities.trunc(self.sourceNormCoeffs[d.name][0], 1)+" std->"+utilities.trunc(self.sourceNormCoeffs[d.name][1], 3)+"", printLevel)
#		p.post("", printLevel)



################################################################################
################################################################################
################################################################################
################################################################################
class segmentedDescriptors:
	def __init__(self, sfhandle, v=False):
		self.sfhandle = sfhandle
		self.calced = {}
		self.calcednorm = {}
		self.dname2d = {}
		self.v=v
	##########
	def addDescriptor(self, dobj):
		self.dname2d[dobj.name] = dobj
	##########
	def setNormCoeffs(self, n1, n2):
		self.n1 = n1
		self.n2 = n2
	##########
	def getraw(self, dname, start, end):
		key = (dname, start, end)
		if not self.calced.has_key(key):
			assert type(key) == tuple
			assert len(key) == 3
			dobj = self.dname2d[key[0]]
			self.calced[key] = metrics.analSegment(dobj, self.sfhandle, key[1], key[2], self.sfhandle.typed, ag)[0]
			#print "eval segment: %s from %s-%s == %.3f"%(key[0], key[1], key[2], self.calced[key])
		return self.calced[key]
	##########
	def getnorm(self, dname, start, end):
		key = (dname, start, end)
		if not self.calcednorm.has_key(key):
			self.calcednorm[key] = (self.getraw(dname, start, end)-self.n1)/self.n2
			if self.v: print "NORM segment: %s from %s-%s == %.3f"%(key[0], key[1], key[2], self.calcednorm[key])
		return self.calcednorm[key]
	##########
	def rawvalues(self, dname):
		'''gives back all calculated raw values of a 
		particular descriptor.  useful for obtaining
		normailization coefficients.'''
		output = []
		dobj = self.dname2d[dname]
		for key in self.calced:
			if key[0] == dname: output.append(self.calced[key])
		return output
	##########
	def clearrange(self, st, ed):
		'''clears out calculated values for both raw
		and normalized dicts over a start and end frame time.
		useful for when you're changed a descriptor (subtracted
		power, for instance) and need new segment estimates.'''
		delkey = []
		for tdname, tst, ted in self.calced:
			if tst <= ed and ted >= st:
				delkey.append((tdname, tst, ted))
		if self.v: print "CLEAR RAW", delkey
		for k in delkey:
			del self.calced[k]
		delkey = []
		for tdname, tst, ted in self.calcednorm:
			if tst <= ed and ted >= st:
				delkey.append((tdname, tst, ted))
		if self.v: print "CLEAR NORM", delkey
		for k in delkey:
			del self.calcednorm[k]



################################################################################
################################################################################
################################################################################
################################################################################
class targetData:
	def __init__(self, filename, startSec, endSec, thresh, rise, scaleDb, minSegLen, maxSegLen, svpbin):
		if not os.path.exists(filename):
			print "\n\nTarget file", filename, "not found!\n"
			utilities.exit(1)
		
		#######################
		## time stretching ? ##
		#######################
		if ops.TARGET_STRETCH_TIME not in [1, None]:
			tmpfile = open('/tmp/stretchtmpfile.txt', 'w')
			tmpfile.write("-10 1\n0 %s\n"%(ops.TARGET_STRETCH_TIME))
			tmpfile.close()
			outfile = '/tmp/target_stretch.aiff'
			if startSec != None:
				ops.TARGET_STRETCH_FLAGS += ' -B%f'%startSec
				startSec = None
			if endSec != None:
				ops.TARGET_STRETCH_FLAGS += ' -E%f'%endSec #If end  contains a decimal  point or ends  with the letter s  it is interpreted as time in seconds,
				endSec = None
			command = '%s -t -Z -U -S"%s" %s -D"/tmp/stretchtmpfile.txt" "%s"'%(svpbin, filename, ops.TARGET_STRETCH_FLAGS, outfile)
			print 'TARGET TIME STRETCH - "%s"'%command
			utilities.executeCommand( command )
			filename = outfile

		
		self.filename = filename
		self.filehead = os.path.split(self.filename)[1]
		self.name = filename.split('/')[-1] # short name for printing
		self.typed = 'target'
		self.startSec = startSec
		self.endSec = endSec
		self.thresh = thresh
		self.rise_ratio_offset = rise
		ag.minSegLength = ag.s2f(minSegLen)
		ag.maxSegLength = ag.s2f(maxSegLen)
		#
		self.sfdur, self.sfchannels = SdifInterface.validateSdifResource( self.filename )
		if self.startSec == None: self.startSec = 0
		if self.endSec == None: self.endSec = self.sfdur
		self.onsetTime = 0
		self.offsetTime = 0
		self.envelopeSlope = 1
		self.scaleDb = scaleDb
		if self.scaleDb == 0: self.envelopeMask = None
		else: self.envelopeMask = utilities.dbToAmp( self.scaleDb )
		ag.allHandles.append(self) # add to all list
		self.data = {'raw': {}, 'norm': {}, 'mix-raw': {}}
		self.segdata = segmentedDescriptors(self)
		# trackers
		self.timeInFrames = 0 # timekeeper
		self.segmentCounter = 0 # count target segments
		self.lastTgtSegTimes = (None, None)
		self.lastNoteTimeInFrames = 0 # for segment tracking
		self.needMcffRecalc = False # hacky
		self.lastMarkovIntMapChoice = None
		self.lastMarkovPitchChoice = None
		self.lastSegmenthLength = 1 # always start at 1
		self.incSmearAdditionalNotes = 0
		self.lastSegmentInitiated = []
		self.lastSegmentTerminated = []
		self.selectedSegs = []
	############################################################################
	def rewind(self):
		# called it tgt is used for another concate..
		self.timeInFrames = 0
		self.segmentsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.onsetsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.data['raw'] = deepcopy(self.data['orig']) # restore unsubtracted/re-analized target data....
		self.segdata = {}
		self.onsetInit()
		src.fileSelectedList = []
		src.nameSelectedList = []
		src.simSelectedList = []
		self.initSegMethod(descriptorData.target_seg)
		self.lastMarkovIntMapChoice = None
		self.lastMarkovPitchChoice = None
		src.timesVoiceWasPlayed = {}
		ag.selectedSegmentTimes = []
		ag.selectedSegments = [] # maintain a list of selected corpus segments
		ag.countRanOutOfSamples = 0
		self.initMixture()
	############################################################################
	def initAnal(self, descriptorsHandle):
		p.post(self.name, inc=True)
		self.startSec, self.endSec, self.startFrame, self.lengthInFrames = SdifInterface.getDescriptorSegment(self.filename, self.startSec, self.endSec, descriptorsHandle.target_load, self.data['raw'], self.envelopeMask)
#		print self.startSec, self.endSec, self.startFrame, self.lengthInFrames
#		print self.startSec, self.endSec, self.startFrame, self.lengthInFrames
#		print self.startSec, self.endSec, self.startFrame, self.lengthInFrames
#		print self.startSec, self.endSec, self.startFrame, self.lengthInFrames
#		print descriptorsHandle.target_seg
#		print descriptorsHandle.target_onfly
#		print descriptorsHandle.target_onfly
#		print descriptorsHandle.target_onfly
#		print descriptorsHandle.target_onfly
		# modifications
		for d in descriptorsHandle.target_load:
			for mathsign, value in d.tgt_modify:
				self.data['raw'][d.name] = eval("self.data['raw']['%s'] %s %f"%(d.name, mathsign, value))

		for d in descriptorsHandle.target_seg:
			#self.data['raw'][d.name] = [None]*self.lengthInFrames # make a tmp data track
			self.segdata.addDescriptor(d)

			
			#self.data['raw'][d.name] = segmentedDescriptor(self, self.data['raw'], self.data['raw']['power'], d)
		for d in descriptorsHandle.target_onfly:
			self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
		self.segmentsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.onsetsPerFrame = np.zeros(self.lengthInFrames, dtype=int) # notekeeper
		self.maxPower = np.max(self.data['raw']['power'])
		self.minPower = np.min(self.data['raw']['power'])
	############################################################################
	def onsetInit(self):
		'''this initializes data for triggerings methods.'''
		self.triggerData = {'totalWeights': 0}
		for d in descriptorData.master_list:
			if d.origin != 'TARGET_ONSET_DESCRIPTORS': continue
			if d.weight <= 0: continue
			self.triggerData[d.name] = {}
			if self.data['raw'].has_key(d.name) == False: self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
			if ops.TARGET_ONSET_FORCE_MAX.has_key(d.name): # hacked to force max val
				self.triggerData[d.name]['initMax'] = ops.TARGET_ONSET_FORCE_MAX[d.name]
			else:
				self.triggerData[d.name]['initMax'] = max(self.data['raw'][d.name])

			self.triggerData[d.name]['initMin'] = min(self.data['raw'][d.name])
			self.triggerData['totalWeights'] += d.weight
			if d.name.find('power') != -1 or d.name.find('rms') != -1 or d.name.find('flux') != -1:
				self.triggerData[d.name]['type'] = 'power'
			else:
				self.triggerData[d.name]['type'] = 'metric'
		self.triggerScalar = self.thresh
		# for simultanous selection amplitude scaling...
		self.triggerThreshNumbNotes = {}
		self.triggerThreshActiveNumbNotes = {} # a higher threshold for subtraced areas?
		# different thresh if there has been a subtraction?
		if ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY != None:
			self.triggerScalarActive = utilities.dbToAmp(ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY)
		else: self.triggerScalarActive = 1
	############################################################################
	def offsetInit(self): 
		'''this initializes data for segment offsets.'''
		minPower = min(self.data['raw']['power'])
		ag.absThreshScalar = utilities.dbToAmp(ops.TARGET_SEGMENT_OFFSET_DB_ABS_THRESH)
		if minPower < ag.absThreshScalar: # use absolute threshold
			ag.powerOffsetValue = ag.absThreshScalar
			p.write("TARGET SEGMENTATION: using an offset amplitude value of %s"%(ag.powerOffsetValue))
		else:
			ag.powerOffsetValue = minPower*utilities.dbToAmp(ops.TARGET_SEGMENT_OFFSET_DB_REL_THRESH)
			p.write("TARGET SEGMENTATION: Warning: the amplitude of %s never got below the offset threshold of %sdB specified in TARGET_SEGMENT_OFFSET_DB_ABS_THRESH.  So, I'm using TARGET_SEGMENT_OFFSET_DB_REL_THRESH dB above the minimum found power -- a value of %f dB."%(tgt.name, ops.TARGET_SEGMENT_OFFSET_DB_ABS_THRESH, utilities.ampToDb(ag.powerOffsetValue)))
	############################################################################
	def thresholdTest(self, time):
		'''Test the target frame TIME to see if
		it is above the triggering threshold.'''
		
		if not self.triggerThreshNumbNotes.has_key(self.onsetsPerFrame[time]):
			self.triggerThreshNumbNotes[self.onsetsPerFrame[time]] = utilities.dbToAmp(self.triggerScalar+(ops.TARGET_ONSET_THRESHOLD_SIM_PENALITY*self.onsetsPerFrame[time]))*self.triggerData['totalWeights']

		
		thisNoteTrigger = self.triggerThreshNumbNotes[self.onsetsPerFrame[time]]
		if ops.TARGET_ONSET_THRESHOLD_ACTIVE_PENALITY != None and len(ag.activeSegments) > 1:
			thisNoteTrigger *= self.triggerScalarActive
			
		triggerTotal = 0 # anything above 1 triggers
		for anal, weight in ops.TARGET_ONSET_DESCRIPTORS.iteritems(): # make sure metrics are analyzed...
			if weight <= 0: continue # skip zeros
			# use below if it is a power triggering method -- ignore negative values...
			if self.triggerData[anal]['type'] == 'power':
				#print anal, time, len(self.data[ops.TARGET_ONSET_ENVELOPE][anal])
				triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMax']
			# else take the absolute value and scale according to 
			elif self.triggerData[anal]['type'] == 'metric':
				if self.data[ops.TARGET_ONSET_ENVELOPE][anal][self.timeInFrames] >= 0:
					triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMax']
				else: # if negative...
					triggerAdd = (self.data[ops.TARGET_ONSET_ENVELOPE][anal][time])/self.triggerData[anal]['initMin']
			triggerTotal += triggerAdd*weight
		if triggerTotal >= thisNoteTrigger: # trigger it?
			self.lastSegmentInitiated = ['thresh', utilities.trunc(utilities.ampToDb(triggerTotal), 1), utilities.trunc(utilities.ampToDb(thisNoteTrigger), 1)]
			return True
		else:
			return False
	############################################################################
	def findRawSegmentLength(self, time):
		#if self.data['raw']['effDur-seg'][time] == None:
		#	ag.segAnal(self, descriptorData.getDescriptorObjByName('effDur-seg'), time, None)
		return self.segdata.getraw('effDur-seg', time, None)
		#return self.data['raw']['effDur-seg'][(time, None)]
	############################################################################
	def initSegMethod(self, descriptors): # get target triggers all at once for segment calc and normalization
		if ag.targetSegmentsFromFile:
			for fi,fiend in tgt.segFrames:
				for d in descriptorData.target_seg:
					self.data['raw'][d.name]
					#ag.segAnal(self, d, fi, fiend)
				#print self.data['tmpraw'][d.name].calced.values()
		else:
			for fi in range(0, tgt.lengthInFrames-1):
				if self.thresholdTest(fi): # then its a hit
					segmentLength = self.findRawSegmentLength(fi)
					for d in descriptorData.target_seg:
						self.segdata.getraw(d.name, fi, fi+segmentLength)
						#ag.segAnal(self, d, fi, fi+segmentLength)
			#print d.name, self.data['tmpraw'][d.name].calced

	############################################################################
	def getSegs(self):
		segs = []
		self.segFrames = [] # an internal list of segments in frame-time
		ag.targetSegmentsFromFile = False # just to keep track
		self.segFrameOnsets = []
		if ops.TARGET_LOAD_SEGMENTS_FILEPATH not in [False, None]:
			# load target segments from a file
			ag.targetSegmentsFromFile = True
			times = utilities.readAudacityLabelFile(ops.TARGET_LOAD_SEGMENTS_FILEPATH, False, False)
			for i in range(len(times)):
				# test to make sure the segments are within start and end second
				if times[i][0] < self.startSec or times[i][1] > self.endSec: continue
				segmentStart = times[i][0]-self.startSec
				segmentLengthFrames = times[i][1]-segmentStart
				segs.append((segmentStart, segmentLengthFrames))
				self.segFrames.append( (ag.s2f(segmentStart), ag.s2f(float(times[i][1])) ) )
				self.segFrameOnsets.append(ag.s2f(segmentStart))
			if len(segs) == 0:
				utilities.error("\n\n\nNo segments FOUND!?!?!?!\n\n")
			return segs
		# else analyze segments on the fly (the default)
		fi = 0
		while fi < self.lengthInFrames-1:
			if self.thresholdTest(fi): # then its a hit
				segmentLengthFrames = self.findRawSegmentLength(fi)
				segmentStart = ag.f2s(fi)
				segs.append((segmentStart, ag.f2s(segmentLengthFrames), self.lastSegmentInitiated, self.lastSegmentTerminated))
				self.segFrames.append((fi, fi+segmentLengthFrames))
				fi += segmentLengthFrames
			else:
				fi += 1
		return segs
	############################################################################
	def writeSegLabelsToDisk(self, path, segs, tempo):
		scale = 60.0/float(tempo)
		segs.sort()
		labels = open(path, "w")
		for i, seg in enumerate(segs):
			writeStr = str(seg[0]*scale)+'\t'+str((seg[0]+seg[1])*scale)
			if seg != None and len(seg) > 2 and len(seg[2]) > 1:
				writeStr += "\tonset "+seg[2][1] # add segmentation reason
			if seg != None and len(seg) > 2 and len(seg[3]) > 1:
				writeStr += " offset "+seg[3][0]+' '+seg[3][1] # add segmentation reason
			labels.write(writeStr+"\n")
		labels.close()
	############################################################################
	def writeOMScore(self, path, segs):
		#frq amp start length channel
		startMSList = ""
		durMSList = ""
		midiAmpList = ""
		midiCentList = ""
		chanList = ""
		for seg in segs:
			startMSList += str(int(seg[0]*1000))+" "
			durMSList += str(int(seg[1]*1000))+" "
			midiAmpList += str(127)+" " # max amp
			midiCentList += str(6000)+" " # middle C
			chanList += str(0)+" "
		OmScoreFile = open(path, 'w')
		OmScoreFile.write("("+midiCentList+") ("+startMSList+") ("+durMSList+") ("+midiAmpList+") ("+chanList+")")
		OmScoreFile.close()
	############################################################################
	def writeSegSfsToDisk(self, segs):
		for i, seg in enumerate(segs):
			print "writing seg sf to disk:", i, '->', seg
			utilities.executeCommand(['sox', '"'+self.filename,os.path.join(ops.SCRIPT_PATH, "output/segs", 'tgt'+str(i)+'.aiff')+'"', 'trim', str(seg[0]), str(seg[1])])
	############################################################################
	def plot(self, normalize=True, dataNames=['centroid']):
		normalize = False
		plotFileName = os.path.join(ops.SCRIPT_PATH, "output/segs", 'plot.png')
		dataArrays = []
		for d in dataNames: # normalize
			listy = np.array(self.data['raw'][d])
			if normalize:
				miny, maxy = min(self.data['raw'][d]), max(self.data['raw'][d])
				listy = (listy-miny)/(maxy-miny)
			dataArrays.append(listy)
		plot(plotFileName, dataArrays, dataNames, segs=self.segFrames, height=3.5)
		#waterfallPlot(os.path.join(ops.SCRIPT_PATH, "output/segs", 'mels'+str(self.timeInFrames)+'.png'), self.data['raw']['melamps'])
	############################################################################
	def featureModification(self, start, end, dataSlot):
		for [d, modType, value] in descriptorData.target_feature_mod:
			self.calcFeatureWarp(d.name, value, modType, dataSlot, start, end)
	############################################################################
	def calcFeatureWarp(self, metric, value, modType, dataSlot, start, end):
		# called by featureModification for each anal...
		value = float(value)
		if start == None and end == None: # this is an init pass...
			#print ""
			#print "Modifying Target post-normalization -> "+metric+modType+str(value)
			#print "\twas:", self.data['norm'][metric][0:5]
			for idx, val in enumerate(self.data[dataSlot][metric]):
				if val == None: continue # skip Nones possible with seg anals
				calcRollOffDataPoint(metric, dataSlot, idx, val, value, modType)
			#print "\tnow:", self.data['norm'][metric][0:5]
		else: # only do a slice...
			for idx, val in enumerate(self.data[dataSlot][metric][start:end]):
				if val == None: continue # skip Nones possible with seg anals
				calcRollOffDataPoint(metric, dataSlot, start+idx, val, value, modType)
	############################################################################
	def onFlySegmentCalc(self, d, start, end, dataType):
		if type(d) == str: d = descriptorData.getDescriptorObjByName(d)
		# returns segment calculation
		if dataType == 'norm':
			return self.segdata.getnorm(d.name, start, end)	
		elif dataType == 'raw':
			return self.segdata.getraw(d.name, start, end)	



#		if dataType == 'norm':
#			if self.data['norm'][d.name][start] == None: # if segment calc for this target frame has not been computed...
#				#ag.segAnal(self, d, start, end)
#				self.data['raw'][d.name][(start, end)]
#				ag.normalize([self], [d.name], 'raw', 'norm', ops.NORMALIZE_METHOD, ag.targetNormCoeffs, start, start+1)
#			return self.data['norm'][d.name][start]
#		elif dataType in ['raw']:
#			if self.data['raw'][d.name][start] == None: # if segment calc for this target frame has not been computed...
#				ag.segAnal(self, d, start, end)
#				
#			return self.data['raw'][d.name][start]
	############################################################################
	def leastSqrsAmp(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		tgtFrames = self.data['raw']['power'][subtractTime:subtractTime+minFrames]
		srcFrames = sampleHandle.data['raw']['power'][:minFrames]#*utilties.dbToAmp(sampleHandle.scaleDb)
		try:
			leastSqrWholeLine = (np.linalg.lstsq(np.vstack([srcFrames]).T, np.vstack([tgtFrames]).T)[0][0][0])
		except np.linalg.linalg.LinAlgError: # in case of incompatible dimensions
			leastSqrWholeLine = 0
			pass
		sourceAmpScale = utilities.limit(leastSqrWholeLine, utilities.dbToAmp(minGain), utilities.dbToAmp(maxGain))
		return sourceAmpScale, 100+utilities.ampToDb(sourceAmpScale)
	############################################################################
	def peakAmpCorrect(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		tgtPower = self.segdata.getraw(ops.POST_SELECTION_AMP_PEAK_AMP_METRIC, subtractTime, subtractTime+subtractDuration)
		srcPower = sampleHandle.segdata.getraw(ops.POST_SELECTION_AMP_PEAK_AMP_METRIC, 0, subtractDuration)
		scale = tgtPower/srcPower
		return scale, 100+utilities.ampToDb(scale)
	############################################################################
	def averagedAmpCorrect(self, sampleHandle, minGain, maxGain, subtractTime, subtractDuration):
		'''averages the scalar difference of the loudest 50% of target frames!'''
		percent = 50.
		minFrames = min([subtractDuration, self.lengthInFrames-self.timeInFrames, sampleHandle.lengthInFrames])
		evalFrame = int((percent/100.)*minFrames)
		tgtFrames = self.data['raw']['power'][subtractTime:subtractTime+minFrames]
		srcFrames = sampleHandle.data['raw']['power'][:minFrames]
		tmp = []
		for fidx, frame in enumerate(tgtFrames):
			tmp.append((frame, fidx))
		tmp.sort()
		tmp.reverse()
		f = 0
		avgs = []
		while f <= evalFrame:
			tFrame, fidx = tmp[f]
			avgs.append(tFrame/srcFrames[fidx])
			f += 1
		scale = np.average(avgs)
		return scale, 100+utilities.ampToDb(scale)
	############################################################################
	def initMixture(self):
		self.data['mix-raw'] = {}
		for d in descriptorData.target_descriptor_mixture:
			if d.seg:
				self.data['mix-raw'][d.name] = {}
			else:
				self.data['mix-raw'][d.name] = np.zeros(self.lengthInFrames)
	############################################################################
	def mixSelectedSamplesDescriptors(self, sampleHandle, sourceAmpScale, mixTime, mixDuration):
		#print "\nDESCRIPTOR MIX", sampleHandle.name, "\ttarget time:", mixTime, "\tmix duration:", mixDuration
		# catch if target requested too far:
		if mixTime+mixDuration > self.lengthInFrames: mixDuration = self.lengthInFrames-mixTime # if too long for tgt
		if mixDuration > sampleHandle.lengthInFrames: mixDuration = sampleHandle.lengthInFrames # if too long for corpus item
		for d in descriptorData.target_descriptor_mixture:
			if not d.seg:
				self.data['mix-raw'][d.name][mixTime:mixTime+mixDuration] =  metrics.mixTwoTimeVaryingFeatures(d.name, mixTime, mixDuration, self.data['mix-raw'], sampleHandle.data['raw'])
			else:
				ag.clearSegDict(self.data['mix-raw'][d.name], mixTime, mixDuration)
	############################################################################
	def subtractPowerForOnsetEnvelope(self, sampleHandle, sourceAmpScale, subScale, start, length):
		length = min(self.lengthInFrames-start, length, sampleHandle.lengthInFrames)
		samplePowers = sampleHandle.data['raw']['power'][0:length]
		rawSubtraction = self.data['raw']['power'][start:start+length]-(samplePowers*sourceAmpScale*subScale)
		# clip it so its above zero
		self.data['raw']['power'][start:start+length] = np.clip(rawSubtraction, 0, sys.maxint)
		# recalculate onset envelope
		for d in descriptorData.target_descriptor_subtract:
			if d.name == 'power': continue # its already been done
			self.data['raw'][d.name][start:start+length] = metrics.analSegment(d, self, start, start+length, "Recalc", ag)
		# erase segs..
		tgt.segdata.clearrange(start, start+length)




################################################################################
################################################################################
class TgtSegment:
	def __init__(self, start, end):
		self.start = start
		self.end = end
		self.power = tgt.segdata.getraw('power-seg', start, end)		
		self.peak = metrics.peakTimeSeg(tgt.data[ops.CONCATENATE_PEAK_ALIGN_ENVELOPE]['power'][start:end])[0]







################################################################################
################################################################################
class SuperimposeTracker():
	def __init__(self, tgtlength, tgtlengthsegs, overlap_inc_amp, cpsentrylength, log):
		self.log = log
		self.cnt = {'onset': np.zeros(tgtlength), 'overlap': np.zeros(tgtlength), 'segidx': np.zeros(tgtlengthsegs), 'cpsvc_overlap': np.zeros((cpsentrylength, tgtlength))}
		self.overlap_inc_amp = utilities.dbToAmp(overlap_inc_amp)
		self.histogram = {'select': {}, 'skip': {}}
	########################################
	def test(self, type, time, min, max):
		pick = 'ok' # 'ok', 'notok', 'force'
		if min != None and self.cnt[type][time] < min: pick = 'force'
		if max != None and self.cnt[type][time] >= max: pick = 'notok'
		return pick	
	########################################
	def increment(self, start, dur, segidx, cps_voiceid, powers):
		self.cnt['segidx'][segidx] += 1
		self.cnt['onset'][start] += 1
		for f in range(dur):
			try:
				if powers[f] >= self.overlap_inc_amp:
					self.cnt['overlap'][start+f] += 1
					self.cnt['cpsvc_overlap'][cps_voiceid][start+f] += 1
			except IndexError: break # cps handle data not long enough
	########################################
	def skip(self, reason, value):
		if not self.histogram['skip'].has_key(reason):
			self.histogram['skip'][reason] = []
		self.histogram['skip'][reason].append(value)
		self.log.write( "SKIP @ %.2f -- %s (%s)\n"%(ag.f2s(tgt.timeInFrames), reason, value) )
	########################################
	def pick(self, trig, trigVal, onsett, overt, segidxt):
		if not trig and onsett == 'force':
			self.log.write( "SELECT @ %.2f -- target too soft but forced to by minOnset\n"%(ag.f2s(tgt.timeInFrames)) )
		elif not trig and overt == 'force':
			self.log.write( "SELECT @ %.2f -- target too soft but forced to by minOverlap\n"%(ag.f2s(tgt.timeInFrames)) )
		elif not trig and segidxt == 'force':
			self.log.write( "SELECT @ %.2f -- target too soft but forced to by minSegment\n"%(ag.f2s(tgt.timeInFrames)) )
		else:
			self.log.write( "SELECT @ %.2f\n"%(ag.f2s(tgt.timeInFrames)) )






	
################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class addSfSegment:
	def __init__(self, filename):
		self.filename = filename
		self.typed = 'source'
		# MANY other attributes set by source calling class
	############################################################################
	def initAnal(self, descriptorsHandle):
		self.sfdur, self.sfchannels = SdifInterface.validateSdifResource( self.filename )
		self.name = self.filename.split('/')[-1] # short name for printing
		self.getOnsetAndOffsetTimes(self.onsetLen, self.offsetLen, self.envelopeSlope)
		# print it
		maxlength = 30
		if len(self.filehead) > maxlength:
			fileheadprint = self.filehead[0:maxlength-1]+' '*5
		else:
			fileheadprint = self.filehead	+ (len(self.filehead)-maxlength+5)*' '
		p.post("%s@ %.2f-%.2f"%(fileheadprint, self.startSec, self.endSec), inc=True)
		self.data = {'raw': {}, 'norm': {}}
		self.segdata = segmentedDescriptors(self)
		self.passedLimitTest = True
		load = []
		onfly = []
		for d in descriptorData.master_list:
			if 'corpus' not in d.neededBy: continue
			if d.type == 'rawsdif': load.append(d)
			else: onfly.append(d)
		self.startSec, self.endSec, self.startFrame, self.lengthInFrames = SdifInterface.getDescriptorSegment(self.filename, self.startSec, self.endSec, load, self.data['raw'], self.envelopeMask)
		for d in onfly:
			if d.seg:
				#self.data['raw'][d.name] = segmentedDescriptor(self, self.data['raw'], self.data['raw']['power'], d)
				self.segdata.addDescriptor(d)
			else:
				self.data['raw'][d.name] = metrics.analSegment(d, self, None, None, "Calc", ag)
	############################################################################
	def getOnsetAndOffsetTimes(self, onset, offset, slope):
		if self.startSec == None: self.startSec = 0
		if self.endSec == None: self.endSec = self.sfdur
		noteDuration = self.endSec-self.startSec
		self.lengthInFrames = SdifInterface.sec2framelength(noteDuration, self.filename)
		cushion = 0.01
		if type(onset) == str and onset.find('%') != -1:
			self.onsetTime = noteDuration*(float(onset.replace('%',''))/100.)
		elif type(onset) in [float, int]:
			self.onsetTime = onset
		else:
			print "ERROR IN ONSET TIME "
			sys.exit(1)

		if type(offset) == str and offset.find('%') != -1:
			self.offsetTime = noteDuration*(float(offset.replace('%',''))/100.)
		elif type(onset) in [float, int]:
			self.offsetTime = offset
		else:
			print "ERROR IN ONSET TIME "
			sys.exit(1)
		if (self.onsetTime+self.offsetTime+cushion) > noteDuration:
			self.offsetTime = noteDuration-self.onsetTime-cushion
		# if SdifLinkage will factor in the envelope applied to this segment..
		onsetFrames = int(round(ag.s2f(self.onsetTime)))
		offsetFrames = int(round(ag.s2f(self.offsetTime)))

		if onsetFrames <= 1 and offsetFrames <= 1:
			self.envelopeMask = utilities.dbToAmp(self.scaleDb)
		else:
			self.envelopeMask = np.ones(self.lengthInFrames)
			if onsetFrames > 0:
				self.envelopeMask[:onsetFrames] = np.linspace(1/float(onsetFrames), 1, num=onsetFrames)
			if offsetFrames > 0:
				self.envelopeMask[offsetFrames*-1:] = np.linspace(1, 1/float(offsetFrames), num=offsetFrames)
			self.envelopeMask = np.power(self.envelopeMask, self.envelopeSlope)
			self.envelopeMask *= utilities.dbToAmp(self.scaleDb)
	############################################################################
	def limitPass(self, limitMaster):
		passed = True
		for (dname, sign, cmpval, cpsId), cnt in limitMaster.iteritems():
			if cpsId != self.voiceID: continue
			if not eval('%f %s %f'%(self.data['raw'][dname][0], sign, cmpval)): 
				passed = False
				cnt += 1
		return passed
			










########################################################
class otherSavedData():
	def __init__(self, fullPath, name):
		# using json
		self.fullPath = fullPath
		self.handle = None
		self.needToWrite = False
		self.triedToRead = False
	####################################################
	def hasData(self, hash):
		# open the file and read if needed...
		if not self.triedToRead:
			if os.path.isfile(self.fullPath): 
				print "Reading json archive", self.fullPath
				self.handle = open(self.fullPath, 'r')
				self.data = json.load(self.handle)
				self.handle.close()
				self.handle = None
			else:
				self.data = {}
			self.triedToRead = True
		return self.data.has_key(hash)
	####################################################
	def getData(self, hash): return self.data[hash]
	####################################################
	def commitData(self, hash, data):
		self.data[hash] = data
		#print self.data[hash]
		self.needToWrite = True
	####################################################
	def close(self):
		if self.needToWrite:
			self.handle = open(self.fullPath, 'w')
			print "writing", self.handle
			json.dump(self.data, self.handle)
	


################################################################################
################################################################################
################################################################################
################################################################################
################################################################################
class outputData:
	def __init__(self, scoreNameOrNone, qlistNameOrNone, tgt, csoundFileName, OmNameOrNone, midifileOrNone):
		self.totalOutputNotes = 0
		self.scoreNotes = []
		self.scoreName = scoreNameOrNone
		self.lastNoteTime = 0.0
		self.lastNoteEndTime = 0.0 #not needed anymore
		if qlistNameOrNone != None: self.qlistHandle = open(qlistNameOrNone, 'w')
		if OmNameOrNone != None: self.OmScoreFile = open(OmNameOrNone, 'w')
		self.midifileOrNone = midifileOrNone
		self.csoundFileName = csoundFileName
		self.scoreMixHandle = tempfile.NamedTemporaryFile('w')
		self.scoreMixHandleName = self.scoreMixHandle.name
		
		self.tgtScoreSegmentHandle = tempfile.NamedTemporaryFile('w')

		self.qlistNotes = []
		self.OmNotes = []
		self.noteHandles = []
		self.handleToNote = {}
		self.minMaxSearch = []
		for seg in tgt.segs:
			self.tgtScoreSegmentHandle.write('\t'.join(['i1', str(seg[0]), str(seg[1]), '"'+tgt.filename+'"', str(seg[0]), '0.5\n']))
		tgtDur = tgt.endSec-tgt.startSec
		#CSOUND_OUTPUT_MIX_PAN = 'middle' # 'middle' || 'leftRight'
		if ops.CSOUND_OUTPUT_MIX_PAN == 'middle': pan = ('1', '1')
		elif ops.CSOUND_OUTPUT_MIX_PAN == 'leftRight': pan = ('0', '1')
		self.scoreMixHandle.write('\t'.join(['i1', '0', str(tgtDur), '"'+tgt.filename+'"', str(tgt.startSec), pan[0], '\n']))
		self.scoreMixHandle.write('\t'.join(['i1', '0', str(tgtDur+3), '"'+csoundFileName+'"', '0', pan[1], '\n']))
		self.outputLabelData = []
	############################################################################
	def changeOutputScoreName(name):
		self.scoreName = name
	############################################################################
	def pstr(self, val):
		return utilities.trunc(val, ops.OUTPUT_FLOAT_PRECISION) # N decimal places for floats.
	############################################################################
	def addEvent(self, handle, timeSec, durSec, tgtDurSec, outputDb, trans, segCntId, simSelects, stretchCode):
		peakTimeSec = ag.f2s(handle.segdata.getraw('peakTime-seg', 0, None))
		effDurSec = ag.f2s(handle.segdata.getraw('effDur-seg', 0, None))
		rawRms = utilities.dbToAmp(utilities.ampToDb(handle.segdata.getraw('power-seg', 0, None)))
		outputDb += ops.OUTPUT_GAIN+handle.scaleDb
		if ops.CSOUND_SCORE_FILEPATH != None:
			outputFiles.addScoreNote("i1", timeSec, durSec, outputDb, handle, trans, handle.onsetTime, handle.offsetTime, handle.envelopeSlope, peakTimeSec, effDurSec, handle.voiceID, simSelects, tgtDurSec, segCntId, int(stretchCode)) # add note to csound sco data
		if ops.QLIST_FILEPATH != None:
			self.addQlistNote(handle, timeSec, trans, outputDb, effDurSec, handle.onsetTime, handle.offsetTime)

		if ops.SEGMENT_LABELS_FILEPATH != None:
			self.outputLabelData.append([self.pstr(timeSec), self.pstr(timeSec+(handle.endSec-handle.startSec)), handle.name+'@'+self.pstr(handle.startSec)+'-'+self.pstr(handle.endSec), ' '.join(handle.extraMappedData).replace('\n',''), '\n'])
	############################################################################
	def addComment(self, listy):
		if ops.CSOUND_SCORE_FILEPATH != None: self.scoreNotes.append(listy)
	############################################################################
	def addScoreNote(self, instru, startTime, absDur, amp, handle, semitones, onsetTime, offsetTime, envSlope, peakTimeSec, effDur, voiceID, simSelectVoice, tgtDurSec, segCntIdx, stretchCode):
		thisTime = (startTime*(60.0/float(ops.TEMPO_CHANGE)))+ops.TIME_OFFSET
		# get output duration
#		if ops.CSOUND_SCORE_DURATION_TYPE == 'dur': 
#			absDur = handle.endSec-handle.startSec
#		elif ops.CSOUND_SCORE_DURATION_TYPE == 'effDur':
#			absDur = effDur
#		elif ops.CSOUND_SCORE_DURATION_TYPE == 'matchDur':
#			absDur = simCalcDuration*ag.frameSec
		#absDur = handle.endSec-handle.startSec
		semitones = utilities.quantize(semitones, handle.transQuantize)
		trans = 2**(semitones/12.0)
		noteDuration = absDur*(1/trans)
				
		self.handleToNote[handle] = [instru, self.pstr(thisTime), self.pstr(noteDuration), self.pstr(amp), '"'+os.path.abspath(handle.filename)+'"', self.pstr(handle.startSec), self.pstr(trans), self.pstr(utilities.ampToDb(handle.segdata.getraw('power-seg', 0, None))), self.pstr(peakTimeSec), self.pstr(effDur), self.pstr(onsetTime), self.pstr(offsetTime), self.pstr(envSlope), str(int(voiceID)), str(int(simSelectVoice)), self.pstr(tgtDurSec), str(segCntIdx)]

		self.scoreNotes.append([instru, self.pstr(thisTime), self.pstr(noteDuration), self.pstr(amp), '"'+os.path.abspath(handle.filename)+'"', self.pstr(handle.startSec), self.pstr(trans), self.pstr(utilities.ampToDb(handle.segdata.getraw('power-seg', 0, None))), self.pstr(peakTimeSec), self.pstr(effDur), self.pstr(onsetTime), self.pstr(offsetTime), self.pstr(envSlope), str(int(voiceID)), str(int(simSelectVoice)), str(tgtDurSec), str(segCntIdx), str(stretchCode)])
		if handle.segdata.getraw('MIDIPitch-seg', 0, None) != -1:
			thisFrq = handle.segdata.getraw('MIDIPitch-seg', 0, None)+semitones
		elif handle.mappedPitch != -1:
#			print handle.name, handle.mappedPitch, handle.extraMappedData
#			print handle.name, handle.mappedPitch, handle.extraMappedData
#			print handle.name, handle.mappedPitch, handle.extraMappedData
			thisFrq = float(handle.mappedPitch)
		else:
			thisFrq = 60 # default frq of 60 for MIDI if cannot find pitch
		self.OmNotes.append([thisTime, noteDuration, amp+ops.OUTPUT_GAIN, handle.name, thisFrq, utilities.ampToDb(handle.segdata.getraw('power-seg', 0, None)), peakTimeSec, effDur, onsetTime, offsetTime, handle, tgtDurSec])
		self.lastNoteTime = startTime
		self.lastNoteEndTime = startTime+absDur
		self.totalOutputNotes += 1 # increment
		src.fileSelectedList.append(handle.name)
		src.nameSelectedList.append(handle.name)
		self.noteHandles.append([handle, thisTime, amp+ops.OUTPUT_GAIN, semitones])
	############################################################################
	def writeOMScoreNote(self):
		WRITE_REAL_DURATIONS = True
		#frq amp start length channel
		startMSList = ""
		durMSList = ""
		midiAmpList = ""
		midiCentList = ""
		chanList = ""
		self.OmNotes.sort() # fuck this
		notesMaybeQuant = utilities.quantizeNoteSequence(self.OmNotes, ops.RHYTHM_QUANTIZE)
		for (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, tgtDurSec) in notesMaybeQuant:
			startMSList += str(int(thisTime*1000))+" "
			if WRITE_REAL_DURATIONS:
				durMSList += str(int(effDur*1000))+" "
			else:
				durMSList += str(10)+" "
			thisAmp =  utilities.limit((amp-100+(rmsPeak)), -40, 0)
			
			midiAmpList += str((thisAmp*3.175)+127)+" "
			midiCentList += str((MIDIpitch)*100)+" "
			chanList += str(0)+" "
		self.OmScoreFile.write("("+midiCentList+") ("+startMSList+") ("+durMSList+") ("+midiAmpList+") ("+chanList+")")
		self.OmScoreFile.close()
	############################################################################
	def writeLispEvents(self, filename):
		WRITE_REAL_DURATIONS = False # if not, write eff durations
		self.OmNotes.sort()
		handle = open(filename, 'w')
		handle.write('( ')
		for idx, (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, tgtDurSec) in enumerate(self.OmNotes):
			##print (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime)
			handle.write( '('+self.pstr(thisTime)+' '+self.pstr(effDur)+' '+self.pstr(MIDIpitch)+' "'+name+'" '+self.pstr(rmsPeak)+') ')
		handle.write(')')
		handle.close()
	############################################################################
	def writeBachScoreNotes(self, fullPath, sourceHandles, csoundOutputFile):
		instuNames = []
		clefs = []
		keys = []
		midiChns = []
		polyphonys = []
		noteOrganizer = {} # a dict of instrNames with a list of notes grouped by onset time
		# get the order of parts
		partOrder = []
		activeInstrNames = []
		for start, end, handle, amp in ag.selectedSegments:
			if handle.MWinstrName not in activeInstrNames: activeInstrNames.append(handle.MWinstrName)

		for instu, dicty in ops.MW_DEFINE_INSTRUMENTS:
			if instu == 'noName': continue
			if instu not in activeInstrNames: continue # only add instruments that get played
			partOrder.append(instu)
			noteOrganizer[instu] = []
		#'MWtext': 'texty'
		# get part metadata
		outlist = []
		for ins in partOrder:
			clefs.append(str(src.MWData[ins]['clef']))
			keys.append(str(src.MWData[ins]['key']))
			midiChns.append(str(src.MWData[ins]['midiChn']))
			polyphonys.append(src.MWData[ins]['polyphony'])
		commandList = []
		# write header
		commandList.append(['reset', 'bang'])
		commandList.append(['target', tgt.filename, str(tgt.startSec*1000), str(tgt.endSec*1000)])
		#commandList.append(['rendered', csoundOutputFile])
		commandList.append(['both', 'clefs', ' '.join(clefs)])
		commandList.append(['both', 'keys', ' '.join(keys)])
		commandList.append(['both', 'voicespacing', '6', '20 '*len(partOrder), '0'])
		commandList.append(['both', '(slotinfo (3 (name technique) (type text)) (4 (name sfinfo) (type text)) (5 ( name amplitudeEnvelope ) ( type function ) ( range 0. 2. ) ( representation ) ( ysnap ( ) ) ( domain 0. 1. ) ( domainslope 0. ) ( width temporal ) ( singleslotfortiednotes 1 ) ))'])

		# note loop
		self.OmNotes.sort()
		for idx, (thisTime, noteDuration, amp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, tgtDurSec) in enumerate(self.OmNotes):
			if not noteOrganizer.has_key(handle.MWinstrName): continue
			ampShape = list((handle.data['raw']['power'][:]-min(handle.data['raw']['power']))/max(handle.data['raw']['power']))
			ampShapeStr = ''
			timeReduceAmpSlope = 3
			if len(ampShape) > 0:
				for aidx, amp in enumerate(ampShape[::timeReduceAmpSlope]):
					timeZeroToOne = utilities.trunc(aidx/float(len(ampShape)/timeReduceAmpSlope), 2)
					ampShapeStr += '('+timeZeroToOne+' '+utilities.trunc(amp, 2)+' 1) '
				ampShapeStr += ' (1. 0. 1)' # cap the end
			noteDuration = effDur
			if src.MWData[handle.MWinstrName]['shortestDur']: noteDuration = 0.1
			noteOrganizer[handle.MWinstrName].append([thisTime, str(MIDIpitch*100), noteDuration, str(100), handle.MWtext, handle.name+' '+self.pstr(handle.startSec)+'-'+self.pstr(handle.endSec), ampShapeStr])
		##################
		# filter notes to make continuous parts - limiting the polyphony (overlap) of each instrument
		for iidx, ins in enumerate(partOrder):
			noteOrganizer[ins].sort()
			overlapCnt = 1
			lastNoteEnded = -1
			for nidx, note in enumerate(noteOrganizer[ins]):
				if polyphonys[iidx] > 1: continue
				overlappage = lastNoteEnded-note[0]
				#print "#"+str(nidx), "start", note[0], "end", note[0]+note[2], "lastEnd", lastNoteEnded, overlappage # a little extra
				if overlappage > 0: 
					# then modify the duration of the last note
					noteOrganizer[ins][nidx-1][2] -= overlappage
				lastNoteEnded = note[0]+note[2]
		##################
		# change time to MS - both start and summed duration
		for iidx, ins in enumerate(partOrder):
			for nidx, note in enumerate(noteOrganizer[ins]):
				noteOrganizer[ins][nidx][2] = str( int((note[0]+note[2]*1000)) )
				noteOrganizer[ins][nidx][0] = str( int(note[0]*1000) )
		##################
		# write bach text
		for voiceNumb, partName in enumerate(partOrder):
			for note in noteOrganizer[partName]:
				commandList.append(['roll', 'addchord', str(voiceNumb+1), '(', note[0], '(', ' '.join(note[1:4]), '( slots ( 3 "'+note[4]+'" ) ( 4 "'+note[5]+'" ) ( 5 '+note[6]+' ) ) ) )'])
		# quantify commands
		meterStr = ("((("+ops.BACH_METER.replace('/',' ')+") ("+str(ops.BACH_TEMPO)+"))) ")*len(partOrder)
		commandList.append(['meters', meterStr])
		# write file
		commands = open(fullPath, 'w')
		for idx, command in enumerate(commandList):
			commands.write( str(idx)+', '+' '.join(command)+';\n')
		commands.close()
	############################################################################
	def addQlistNote(self, chosenSample, startTime, trans, outputDb, noteDuration, onsetTime, offsetTime):
		thisNoteStart = str(int(chosenSample.startSec*1000))
		thisNoteDur = str(int((chosenSample.endSec)*1000)-int(chosenSample.startSec*1000))
		self.qlistNotes.append([startTime, 'play', chosenSample.filename, thisNoteStart, thisNoteDur, self.pstr(outputDb), self.pstr(trans), self.pstr(int(onsetTime*1000)), self.pstr(int(offsetTime*1000))])
	############################################################################
	def writeQlistNotes(self):
		self.qlistNotes.sort()
		lastNoteTime = 0
		for note in self.qlistNotes:
			delayMS = int((note[0]-lastNoteTime)*1000)
			lastNoteTime = note[0]
			note[0] = str(delayMS)
		outputListOfLists = utilities.printableList(self.qlistNotes)
		for note in outputListOfLists:
			self.qlistHandle.write(note)
		self.qlistHandle.close()
	############################################################################
	def writeScoreNotes(self):
		if ops.ALIGN_NOTE_PEAKS:
			for lidx in range(len(self.scoreNotes)):
				if self.scoreNotes[lidx][0][0] == ';': continue # only a comment
				#print "note items:", self.scoreNotes[lidx]
				newTime = max(0, float(self.scoreNotes[lidx][1])-float(self.scoreNotes[lidx][8]))
				self.scoreNotes[lidx][1] = str(newTime)

		if ops.RHYTHM_QUANTIZE not in [None, 0]: # then quantize the score
			#self.scoreNotes.sort(key=lambda x: x[1])
			notePeaks = []
			for line in self.scoreNotes:
				if line[0][0] == ";": continue # just a comment
				notePeaks.append(float(line[1])+float(line[8]))
			quantizes = sliceListIntoEvenlySpacedLists(notePeaks, ops.RHYTHM_QUANTIZE, 0, max(notePeaks)+1)
			# do histogram to find most common attack point for each window...
			for idx, noteList in enumerate(quantizes): 
				if len(quantizes[idx]) > 0: quantizes[idx] = utilities.histogram(noteList)[0][1] # only if there are notes

			if ops.RHYTHM_QUANTIZE_METHOD == 'aggregate': # align notes in each window to most common attack point...
				for line in self.scoreNotes:
					if line[0][0] == ";": continue # just a comment
					idx = int((float(line[1])+float(line[8]))/ops.RHYTHM_QUANTIZE)
					line[1] = str(quantizes[idx]) # change start time of this note

#			elif ops.RHYTHM_QUANTIZE_METHOD == 'strip': # remove notes not peaking at the most common peak time...
#				for i, line in enumerate(self.scoreNotes):
#					idx = int(notePeaks[i]/ops.RHYTHM_QUANTIZE)
#					peakTime = quantizes[idx]
#					print notePeakTime, peakTime
#					if notePeakTime == peakTime:	
#						self.writeSco('\t'.join(line))
		# sort them in time
		self.scoreNotes.sort(key=lambda x: float(x[1]))
		if self.scoreName != None:
			self.scoreHandle = open(self.scoreName, 'w') # open handle
			self.scoreHandle.write(';'+tgt.filename+' '+self.pstr(tgt.startSec)+' to '+self.pstr(tgt.endSec)+'\n') # write filename to score
			self.scoreHandle.write('i2 0 %f "%s" %f\n'%(tgt.endSec-tgt.startSec, tgt.filename, tgt.startSec))
			
			
			self.scoreHandle.write(';'+"\t".join(['ins', 'start', 'dur', 'amp/100', 'filename', 'start-time', 'trans', 'raw-rms', 'peak-time', 'effDur', 'attack-time', 'decay-time', 'env-slope', 'corpusID', 'simSelects', 'tgtDurSec', 'tgtSegIdx', 'stretchCode'])+'\n') # write comment to score
			outputListOfLists = utilities.printableList(self.scoreNotes)
			for strline in outputListOfLists:
				self.scoreHandle.write(strline)
			self.scoreHandle.close() # close handle
		ag.scoresToRender.append(csoundRenderObj(self.scoreName, ops.CSOUND_RENDER_FILEPATH, outputFiles.totalOutputNotes))

		if ops.MAXMSP_DICT_PATH != None:
			outputLabels = open(ops.MAXMSP_DICT_PATH, 'w')
			outputJsonMaxMsp = {'corpus_files': [], 'target_file': [tgt.filename, tgt.startSec*1000, tgt.endSec*1000], 'events': []}
			for i, st, dur, amp, file, sttime, trans, rms, atk, dec, fadein, fadeout, o1, o2, o3, o4, o5, o6  in self.scoreNotes:
				file = file.replace('"','')
				if file not in outputJsonMaxMsp['corpus_files']: outputJsonMaxMsp['corpus_files'].append(file)
				outputJsonMaxMsp['events'].append([float(st)*1000, float(dur)*1000, file, float(sttime)*1000, float(amp)-100, float(trans), float(fadein)*1000, float(fadeout)*1000])
			json.dump(outputJsonMaxMsp, outputLabels)
			outputLabels.close()
		# creates dicts with voicing ID as key and a list of line numbers as val
		used = {'overlapId': {}, 'corpusId': {}}
		for nidx, noteLine in enumerate(self.scoreNotes):
			if noteLine[0] != 'i1': continue # just a comment
			if noteLine[12] not in used['corpusId']:
				used['corpusId'][noteLine[12]] = []
			used['corpusId'][noteLine[12]].append(nidx)
			if noteLine[13] not in used['overlapId']:
				used['overlapId'][noteLine[13]] = []
			used['overlapId'][noteLine[13]].append(nidx)
			
		if ops.CSOUND_RENDER_SEPARATE_FILE_FOR_EACH_CORPUS_OVERLAP and len(used['overlapId']) > 1:
			for vcId, lines in used['overlapId'].iteritems():
				thisScoName = os.path.join(ops.TMP_PATH, 'overlap'+str(vcId)+'.sco')
				thisRenderOutput = os.path.splitext(ops.CSOUND_RENDER_FILEPATH)[0]+'-overlap'+str(vcId)+os.path.splitext(ops.CSOUND_RENDER_FILEPATH)[1]
				tmpLines = []
				for lidx in lines: tmpLines.append(self.scoreNotes[lidx])
				outputListOfLists = utilities.printableList(tmpLines)
				tmpHandle = open(thisScoName, 'w') # open handle
				for strline in outputListOfLists:
					tmpHandle.write(strline)
				tmpHandle.close() # close handle
				ag.scoresToRender.append(csoundRenderObj(thisScoName, thisRenderOutput, len(outputListOfLists)))

		if ops.CSOUND_RENDER_SEPARATE_FILE_FOR_EACH_CORPUS_ENTRY and len(used['corpusId']) > 1:
			for vcId, lines in used['corpusId'].iteritems():
			
				thisScoName = os.path.join(ops.TMP_PATH, 'corpus'+str(vcId)+'.sco')
				thisRenderOutput = os.path.splitext(ops.CSOUND_RENDER_FILEPATH)[0]+'-corpus'+str(vcId)+os.path.splitext(ops.CSOUND_RENDER_FILEPATH)[1]
				tmpLines = []
				for lidx in lines: tmpLines.append(self.scoreNotes[lidx])
				outputListOfLists = utilities.printableList(tmpLines)
				tmpHandle = open(thisScoName, 'w') # open handle
				for strline in outputListOfLists:
					tmpHandle.write(strline)
				tmpHandle.close() # close handle
				ag.scoresToRender.append(csoundRenderObj(thisScoName, thisRenderOutput, len(outputListOfLists)))
	############################################################################
	def writeParameterScore(self):
		fh = open(ops.PARAM_SCORE_FILEPATH, 'w')
		commentline = ';time file minSearchDist maxSearchDist param1 [param2 param3...]\n'
		fh.write(commentline)
		for (start, stop) in tgt.segFrames:
			scoreline = 'tgt-segment %f %f %f %f %f ; start, stop, peak, power, maxpower\n'%(ag.f2s(start), ag.f2s(stop), ag.f2s(start+tgt.segdata.getraw('peakTime-seg', start, stop)), tgt.segdata.getraw('power-seg', start, stop), tgt.maxPower)
			fh.write(scoreline)
		for t in range(tgt.lengthInFrames):
			scoreline = 'tgt-amp %f %f ; time amp\n'%(ag.f2s(t), tgt.data['orig']['power'][t]/tgt.maxPower)
			fh.write(scoreline)
		for hidx, (handle, thisTime, agAmp, semitones) in enumerate(self.noteHandles):
			if handle.hasParams:
				scoreline = 'segment %f %s %f %f %s'%(thisTime, handle.filename, self.minMaxSearch[hidx][0], self.minMaxSearch[hidx][1], ' '.join(handle.extraMappedData))
				fh.write(scoreline)
			else:
				scoreline = 'corpusSnd %s\n'%' '.join(self.handleToNote[handle][1:])
				fh.write(scoreline)
		fh.close()
	############################################################################
	def done(self, csoundNameOrNone, qlistNameOrNone, OmNameOrNone, midifileOrNone):
		if csoundNameOrNone != None:
			self.filesToRender = self.writeScoreNotes()
		
		if qlistNameOrNone != None: self.writeQlistNotes()
		
		if ops.SEGMENT_LABELS_FILEPATH != None:
			outputLabels = open(ops.SEGMENT_LABELS_FILEPATH, 'w')
			self.outputLabelData.sort()
			for data in self.outputLabelData: outputLabels.write('\t'.join(data))
			outputLabels.close()
			
		self.scoreMixHandle.close()
		if midifileOrNone != None:
			import midifile
			MyMIDI = midifile.MIDIFile(1)
			MyMIDI.addTrackName(0, 0., "AudioGuide Track")
			MyMIDI.addTempo(0, 0., int(ops.TEMPO_CHANGE))
			tempoScaler = float(ops.TEMPO_CHANGE)/60.
			for thisTime, noteDuration, midiAmp, name, MIDIpitch, rmsPeak, peakTimeSec, effDur, onsetTime, offsetTime, handle, tgtDurSec in self.OmNotes:
				midiTime = thisTime*tempoScaler
				midiDuration = noteDuration*tempoScaler
				MyMIDI.addNote(0, 0, MIDIpitch, midiTime, midiDuration, midiAmp)
			
			# And write it to disk.
			binfile = open(self.midifileOrNone, 'wb')
			MyMIDI.writeFile(binfile)
			binfile.close()

		if ops.PARAM_SCORE_FILEPATH != None: self.writeParameterScore()


############################################################################
############################################################################
############################################################################
############################################################################
class csoundRenderObj:
	def __init__(self, scoName, outputName, numberEvents):
		self.scoName = scoName
		self.outputName = outputName
		self.events = numberEvents


def renderWithCsound(renderObjs, barString):
	import subprocess
	totalEvents = 0
	totalEventCounter = 0
	for rObj in renderObjs: totalEvents += rObj.events
	p.barOpen(barString, totalEvents-1)
	for rObj in renderObjs:
		csoundCommand = []
		if ops.CSOUND_RENDER_WITH_ULIMIT:
			csoundCommand.extend(['ulimit -n 2000 ;'])
		csoundCommand.extend(['csound', '-d', '--output="'+rObj.outputName+'"', '"'+os.path.join(ops.AG_PATH,'scripts','csoundRender.orc')+'"', '"'+rObj.scoName+'"' ]	)
		p.write('RENDER WITH CSOUND --> "' + ' '.join(csoundCommand)+ '"\n')
		print ' '.join(csoundCommand)
		cs = subprocess.Popen(' '.join(csoundCommand), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
		while True:
			o = cs.stderr.readline()
			if o == '' and cs.poll() != None: break
			o = o.split()
			if len(o) < 3: continue
			if o[2] == 'giNoteCounter': 
				totalEventCounter += 1
				p.post('Note '+str(totalEventCounter)+'/'+str(totalEvents)+".", inc=True, forceInc=totalEventCounter)
	p.barClose("Done.")



###########################################
###########################################
class cpsLimit:
	def __init__(self, origString, cpsScope):
		self.origString = origString
		limit_pieces = utilities.parseEquationString(origString, ['==', '!=', '<', '<=', '>', '>='])
		self.d = descriptorData.getDescriptorObjByName(limit_pieces[0])
		self.symb = limit_pieces[1]
		# test to see if it is a percentage
		if limit_pieces[2].find('%') == -1:
			self.needMinMax = True
			self.value = float(limit_pieces[2])
		else: # its a percentage
			self.needMinMax = False
			self.percent = float(limit_pieces[2].replace('%', ''))
		self.cpsScope = cpsScope
		self.cnt_reject = []
		ag.limitObjList.append(self)
	########################################
	def checkminMax(self, allcpshandles):
		if self.needMinMax: return
		# loop through corpus handles again, but only happens once per unique limit string
		tmp_data = []
		for ch in allcpshandles:
			if ch.voiceID not in self.cpsScope: continue # skip if outside scope
			if self.d.seg:
				tmp_data.append( ch.segdata.getraw(self.d.name, 0, None) )
			else:
				tmp_data.extend( ch.segdata.getraw(self.d.name, 0, None) )
		tmp_data.sort()
		self.value = tmp_data[ int((self.percent/100.)*(len(tmp_data)-1)) ]
		del tmp_data
		self.needMinMax = True
	########################################
	def test(self, sfobj):
		if self.d.seg:
			test = eval("%s %s %s"%(sfobj.segdata.getraw(self.d.name, 0, None), self.symb, self.value))
		else:
			test = eval("%s %s %s"%(np.max(sfobj.data['raw'][self.d.name]), self.symb, self.value))
		if not test:
			self.cnt_reject.append(sfobj)
		return test
	########################################
	def printRejects(self):
		dicty = {}
		#src.corpusInfoData['filehead']
		for sf in self.cnt_reject:
			id = os.path.split(sf.cpsName)[1]
			if not dicty.has_key(sf.voiceID): dicty[sf.voiceID] = 0
			dicty[sf.voiceID] += 1
		output = self.origString+'\n'
		for voiceID, numb in dicty.iteritems():
			percent = numb*100. / float(src.corpusInfoData[voiceID]['segs'])
			output += "\tremoved %.1f%% segments from %s\n"%(percent, src.corpusInfoData[voiceID]['filehead'])
		return output

###########################################





################################################################################
################################################################################
################################################################################
class source:
	def __init__(self, sourceFromOpts):
		if len(ops.CORPUS) == 0: return # maybe just analyzing a target?
		printVerbose = False
		ag.preLimitCpsHandles = []
		self.corpusInfoData = {}
		self.simSelectRuleByCorpusId = {}
		self.corpusPocketingAssignment = {} # (start, stop): corpusID
		self.lastVoice = None
		# below for histograms
		self.fileSelectedList = []
		self.nameSelectedList = []
		self.voiceRestrictPerFrame = {}
		self.timesVoiceWasPlayed = {}
		self.vcToCorpusName = {}
		self.numberVc = len(ops.CORPUS)
		self.pool = {}
		self.hashtracker = []
		for i in range(self.numberVc): self.pool[i] = {}
		# FOR MUSICAL WRITING
		self.MWData = {} # same as ops.MW_DEFINE_INSTRUMENTS, but with default values added if not supplied
		for MWinstrName, dataDict in ops.MW_DEFINE_INSTRUMENTS:
			self.MWData[MWinstrName] = {'clef': 'G', 'midiChn': 1, 'key': "CM", 'polyphony': sys.maxint, 'shortestDur': False}
			for key, val in dataDict.iteritems():
				self.MWData[MWinstrName][key] = val
		
		globalLimits = []
		if ops.CORPUS_GLOBAL_ATTRIBUTES.has_key('limit'):
			for stringy in ops.CORPUS_GLOBAL_ATTRIBUTES['limit']:
				globalLimits.append( cpsLimit(stringy, range(len(ops.CORPUS))) )


		for cidx, cobj in enumerate(ops.CORPUS):
			# check to make this corpus hasn't already been added
			tmp_items = []
			for name in dir(cobj):
				if name.find('_') != -1: continue
				tmp_items.extend(( name, getattr(cobj, name)))
			hash = utilities.listToCheckSum(tmp_items)
			if hash in self.hashtracker:
				print "\n\nERROR, two identical corpus entries\n\n"
			else:
				self.hashtracker.append(hash)
			# add this voice
			cobj.voiceID = cidx
			if cobj.concatFileName == None: cobj.concatFileName = cobj.name
			self.simSelectRuleByCorpusId[cidx] = cobj.superimposeRule
			self.vcToCorpusName[cidx] = cobj.name

			for name, val in ops.CORPUS_GLOBAL_ATTRIBUTES.iteritems():
				if name == 'limit': continue
				setattr(cobj, name, val)

			# add local limits
			totalLimitList = []
			for limitstr in cobj.limit:
				totalLimitList.append( cpsLimit(limitstr, [cidx]) )
			# add global limits
			totalLimitList.extend(globalLimits) # from CORPUS_GLOBAL_ATTRIBUTES

			# get segments/files list
			timeList = []
			if os.path.isdir(cobj.name): fileType = 'dir'
			if os.path.isfile(cobj.name): fileType = 'file'
			if os.path.islink(cobj.name): fileType = 'link'
			
			# make a log entry
			p.logsection('CORPUS')
			logStr = '* #%i - %s - %s\n'%(cidx, fileType.upper(), cobj.name)
			for name in dir(cobj):
				if name[0] == '_': continue
				if name in ['voiceID', 'concatFileName', 'name']: continue
				if cobj._defaults.has_key(name) and getattr(cobj, name) == cobj._defaults[name] : continue # skip default settings
				val = getattr(cobj, name)
				if type(val) == str:
					logStr += "\t%s = '%s'\n"%(name, getattr(cobj, name))
				else:
					logStr += "\t%s = %s\n"%(name, getattr(cobj, name))
			p.write(logStr)

			
			# use non-standard location segmentation file?
			if cobj.segmentationFile == None: # add default name of segmentation file if nothing is specified by the user			
				cobj.segmentationFile = cobj.name+cobj.segmentationExtension
			elif type(cobj.segmentationFile) == str: # if a single string
				cobj.segmentationFile = cobj.segmentationFile
#			elif type(cobj.segmentationFile) in [tuple, list]: # a list of seg files
#				tmp = []
#				for string in cobj.segmentationFile:
#					tmp.append(path.test(string, ops.CORPUS_SEARCH_PATH)[1])
#				cobj.segmentationFile = tmp
			if not cobj.segmentationFile:
				utilities.error("ops", "Bad Source input name -> %s.  Not a file, a directory"%cobj.segmentationFile)
			
			if fileType == 'file': # an audio file
				##################
				## input a FILE ## -- look for audacity-style txt label file
				##################
				times = []
				#for segFile in cobj.segmentationFile:
				if not cobj.wholeFile and not os.path.isfile(cobj.segmentationFile):
					utilities.error('segmentation file', "Cannot find segmentation file '%s'"%cobj.segmentationFile)
				times.extend(utilities.readAudacityLabelFile(cobj.segmentationFile, cobj.wholeFile, printVerbose))
				for timeSeg in times:
					writeLine = [cobj.name]
					writeLine.extend(timeSeg)
					timeList.append( writeLine )
				cobj.numbSfFiles = 1
			elif fileType == 'dir': # a directory
				#######################
				## input a DIRECTORY ##
				#######################
				files = utilities.getDirListOnlyExt(cobj.name, cobj.recursive, ops.SOUNDFILE_EXTENSIONS)
				cobj.segmentationFile = None # don't print it
				for file in files:
					segFileTest = file+cobj.segmentationExtension
					if not cobj.wholeFile and not os.path.isfile(segFileTest):
						utilities.error('segmentation file', "Cannot find segmentation file '%s'.  To specify the use of whole sound files as corpus segments, write this corpus entry as: \n\tcsf('%s', wholeFile=True)\nor\n\tCORPUS_GLOBAL_ATTRIBUTES = {'wholeFile': True}"%(cobj.segmentationFile, cobj.name))
					times = utilities.readAudacityLabelFile(file+cobj.segmentationExtension, cobj.wholeFile, printVerbose)
					for timeSeg in times:
						writeLine = [file]
						writeLine.extend(timeSeg)
						timeList.append( writeLine )
				cobj.numbSfFiles = len(files)
			# reset counters...
			segCount = 0
			windowDist = metrics.hannWin(len(timeList)*2)
			# restrictInTime if a non-zero positive float/int
			if cobj.restrictInTime > 0:
				self.voiceRestrictPerFrame[cidx] = ag.s2f(cobj.restrictInTime)
			# segment list
			for idx in range(len(timeList)): 
				#thisSegDict = {}
				filehead = os.path.split(timeList[idx][0])[1]
				startSec = timeList[idx][1]
				endSec = timeList[idx][2]
				# matchSting: includeStr/excludeStr
				if cobj.includeStr != None:
					skip = True
					if type(cobj.includeStr) not in [list, tuple]: cobj.includeStr = [cobj.includeStr]
					for test in cobj.includeStr:
						if utilities.matchString(filehead, test, caseSensative=True): skip = False
					if skip: continue
				if cobj.excludeStr != None:
					skip = False
					if type(cobj.excludeStr) not in [list, tuple]: cobj.excludeStr = [cobj.excludeStr]
					for test in cobj.excludeStr:
						if utilities.matchString(filehead, test, caseSensative=True): skip = True
					if skip: continue
				#        minTime / maxTime
				if cobj.start != None and startSec < cobj.start: continue # skip it
				if cobj.end != None and startSec > cobj.end: continue # skip it
				# matchTime: includeTimes/excludeTimes
				if len(cobj.includeTimes) > 0:
					skip = True
					for timeTuple in cobj.includeTimes:
						if startSec >= timeTuple[0] and endSec <= timeTuple[1]: skip = False
					if skip: continue
				if len(cobj.excludeTimes) > 0:
					skip = False
					if type(cobj.excludeTimes) not in [list, tuple]: cobj.excludeTimes = [cobj.excludeTimes] # force it to be a list
					for timeTuple in cobj.excludeTimes:
						if startSec >= timeTuple[0] and endSec <= timeTuple[1]: skip = True
					#print filehead, start, end, skip
					if skip: continue
					
				if len(timeList[idx]) >= 3: extraMappedData = timeList[idx][3:]
				else: extraMappedData = ''
				# test if limitDur is set...
				if cobj.limitDur != None:
					if endSec != None and endSec-startSec > cobj.limitDur: endSec = start+cobj.limitDur
				
				
				# make sfsegment object
				sf = addSfSegment(timeList[idx][0])
				sf.filehead = filehead
				sf.startSec = startSec
				sf.endSec = endSec
				sf.cpsName = cobj.name
				sf.limit = totalLimitList
				sf.segNumber = segCount
				sf.mappedPitch = -1
				sf.extraMappedData = extraMappedData

				for attr in ['MWinstrName', 'MWtext', 'Wnotehead', 'concatFileName', 'hasParams', 'midiPitchMethod', 'numbSfFiles', 'offsetLen', 'onsetLen', 'restrictInTime', 'restrictOverlaps', 'scaleDb', 'scaleDistance', 'superimposeRule', 'transMethod', 'transQuantize', 'voiceID', 'allowRepetition', 'restrictRepetition', 'envelopeSlope']:
					setattr(sf, attr, getattr(cobj, attr))
				
				ag.preLimitCpsHandles.append(sf)
				
				segCount += 1
				###########################
				## done with CORPUS loop ##
				###########################
			self.corpusInfoData[cidx] = {'name': cobj.name, 'filehead': os.path.split(cobj.name)[1], 'segs': str(segCount), 'fileType': fileType, 'limit': totalLimitList, 'limitStart': cobj.start, 'limitEnd': cobj.end, 'numbSfFiles': cobj.numbSfFiles, 'restrictInTime': cobj.restrictInTime, 'segFile': cobj.segmentationFile, 'restrictOverlaps': cobj.restrictOverlaps, 'scaleDb': cobj.scaleDb}		
	############################################################################
	def nameTest(self, name, prefixList): # return only same with a certain starting prefix
		test = True
		name = name.split('/')[-1]
		for t in prefixList:
			if name[0:len(t)] == t:
				test = True
				break
			else:
				test = False
		return test
	############################################################################
	def initConcatenate(self, tgtLength, allSourceHandles):
		for handle in ag.sourceHandles:
			self.pool[handle.voiceID][handle] = 0 # clear time-dependant pool
		# for target seg increment...
		counter = 0
		self.possibleVoicesPerFrame = []
		for i in range(0, tgt.lengthInFrames): self.possibleVoicesPerFrame.append([])
		for pair in tgt.segFrames:
			for i in range(pair[0], pair[1]):
				self.possibleVoicesPerFrame[i].append(counter%src.numberVc)
			counter += 1
		# to count overlaps of each corpus voice
		self.corpusVcOverlapCounter = {}
		for vc in range(self.numberVc): self.corpusVcOverlapCounter[vc] = np.zeros(tgt.lengthInFrames, dtype=int)
	############################################################################
	def updateSourcePool(self, h):
		if h.allowRepetition and h.restrictRepetition != None:
			self.pool[h.voiceID][h] = tgt.timeInFrames+(ag.s2f(h.restrictRepetition))
		elif not h.allowRepetition: 
			del self.pool[h.voiceID][h] # remove it altogether
	############################################################################
	def updateTargetPocketCorpusRestriction(self, tgtTime, tgtSegs, chosenHandle):
		# first check to see if this selection already has a pcoket craved out for it

		for st, ed in self.corpusPocketingAssignment:
			if tgtTime >= st and tgtTime <= ed:
				return
		
		# it does not, so make one!
		for sidx, (st, ed) in enumerate(tgtSegs):
			if tgtTime >= st and tgtTime <= ed:
				break
		pocketStartFrame = tgtSegs[sidx][0]
		numberOfSegmentsToPocket = random.choice(range(int(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[0]), int(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[1])))
		#numberOfSecondsToPocket = random.uniform(ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[0], ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE[1])
		#pocketEndFrame = int(pocketStartFrame+(numberOfSecondsToPocket/ag.frameSec))
		pocketEndFrame = tgtSegs[sidx:sidx+numberOfSegmentsToPocket][-1][1]
		self.corpusPocketingAssignment[(pocketStartFrame, pocketEndFrame)] = chosenHandle.voiceID
		print "VOICE %i solo pocket for %i segments"%(chosenHandle.voiceID, numberOfSegmentsToPocket)
		print '\n\n'


	############################################################################
	def evaluateValidSamples(self, maxActiveVoices, superimp):
		validHandles = []
		validSteams = []
		for vc, pool in self.pool.iteritems():
			# methods for restricting voices based upon the last selected voice
			if src.lastVoice != None:
				# ROTATE_VOICES is a bool
				if ops.ROTATE_VOICES: 
					if vc != (src.lastVoice+1)%src.numberVc: continue # skip this set of samples
				# VOICE_PATTERN
				if ops.VOICE_PATTERN not in [None, []]: # not case sensative!, oes a search, so matches partial strings
					if not utilities.matchString(src.vcToCorpusName[vc], ops.VOICE_PATTERN[len(ag.selectedSegments)%len(ops.VOICE_PATTERN)], caseSensative=False): continue
			# do voice restriction per second
			if src.voiceRestrictPerFrame.has_key(vc):
				srt_look = max(0, tgt.timeInFrames-src.voiceRestrictPerFrame[vc])
				end_look = min(tgt.lengthInFrames, tgt.timeInFrames+src.voiceRestrictPerFrame[vc])
				if np.sum(superimp.cnt['cpsvc_overlap'][vc][srt_look:end_look]) != 0: continue
			# do voice restriction per segment using self.possibleVoicesPerFrame
			if ops.VOICE_RESTRICT_PER_SEGMENT and vc not in self.possibleVoicesPerFrame[tgt.timeInFrames] and len(self.possibleVoicesPerFrame[tgt.timeInFrames]) > 0: continue

			# voice pocketing -- permits only one corpus voice per random grouping of target segments
			#	ops.CORPUS_POCKETING = True
			#	ops.CORPUS_POCKET_SEGMENT_RANDOM_RANGE = [1, 4] # in seconds
			#	ops.CORPUS_POCKET_ONLY_SELECTION_NUMBER = [1, 2, 3] # which sim selections
			#	uses self.corpusPocketingAssignment as a registry to keep track
			#		self.corpusPocketingAssignment = {} # (start, stop): corpusID
			if ops.CORPUS_POCKETING:
				CORPUS_POCKET_WHICH_SIM_SELECTS = [1]
				if tgt.segmentsPerFrame[tgt.timeInFrames] not in CORPUS_POCKET_WHICH_SIM_SELECTS:
					
					prev_pocket_vc = None
					closest_prev_pocket = []
					has_pocket = False
					for (start, stop), vcId in self.corpusPocketingAssignment.iteritems():
						if start <= tgt.timeInFrames and stop >= tgt.timeInFrames:
							has_pocket = True
							
						if stop < tgt.timeInFrames: closest_prev_pocket.append([tgt.timeInFrames-stop, vcId])
					closest_prev_pocket.sort()
					if has_pocket:
						if vcId != vc: 
							continue
					else:
						if ops.CORPUS_POCKETING_ROTATE and len(closest_prev_pocket) > 0:
							if len(closest_prev_pocket) == 1:
								print "ONLY ONE LAST POCKET", closest_prev_pocket[0][1]
								invalidForRotation = [closest_prev_pocket[0][1]]
							else:
								print "TWO LAST POCKETS", closest_prev_pocket[0][1], closest_prev_pocket[1][1], 
								invalidForRotation = [closest_prev_pocket[0][1], closest_prev_pocket[1][1]]
							if vc in invalidForRotation: continue
	
	
				#print "\n\n"
			# restrict corpus ID by number of selected overlapping samples
			# uses eval() to test agains overlap number...
			if self.simSelectRuleByCorpusId.has_key(vc) and self.simSelectRuleByCorpusId[vc] != None:
				if not eval(str(tgt.segmentsPerFrame[tgt.timeInFrames])+self.simSelectRuleByCorpusId[vc]): continue
			###################################################
			## loop through each sound segment in this voice ##
			###################################################
			for sampleHandle, time in pool.iteritems(): # add valid samples to the sample list for this frame
				if sampleHandle.restrictOverlaps != None: # limiting each corpus entry by the number of simultanous overlaps..
					voicesPerFrame = self.corpusVcOverlapCounter[vc][tgt.timeInFrames:tgt.timeInFrames+sampleh.segdata.getraw('effDur-seg', 0, None)]
					#bach noteprint "\n\nVOICES", np.max(voicesPerFrame), "\n\n"
					if np.max(voicesPerFrame) >= sampleHandle.restrictOverlaps: continue
					#print "PASSED", sampleHandle.name, sampleHandle.restrictOverlaps, np.max(voicesPerFrame)
				# if the sample is valid to be picked...
				if time <= tgt.timeInFrames: 
					availableStreams = src.timeRestrictTest(sampleHandle) # tests based upon time restrictions
					if len(availableStreams) > 0:
						validHandles.append(sampleHandle) # add it if you've made it this far
						validSteams.append(availableStreams)
		return validHandles, validSteams
	############################################################################
	def initTimeVaryingRestrict(self):
		if len(ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME) > 0: self.restrictTimeVarying = True
		else: self.restrictTimeVarying = False
		if self.restrictTimeVarying:
			self.sourceTimeRestrictAnal = {}
			for restrict in ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME:
				print "BROKEN, next line should be descriptor object"
				minVal, maxVal = getNormCoefs(ag.sourceHandles, restrict['anal'], 'minmax', None, None)
				self.sourceTimeRestrictAnal[restrict['anal']] = {}
				if restrict.has_key('inits') and len(restrict['inits']) >= ops.CORPUS_RESTRICT_NUMBER_STREAMS:
					initVals = restrict['inits']
				else: 
					initVals = [None]*ops.CORPUS_RESTRICT_NUMBER_STREAMS # no init if none supplied
				for i in range(ops.CORPUS_RESTRICT_NUMBER_STREAMS):
					self.sourceTimeRestrictAnal[restrict['anal']][i] = {'lastActive': 0, 'lastMetircVal': initVals[i], 'minVal': minVal, 'maxVal': maxVal, 'restrictionReleaseSec': restrict['restrictionReleaseSec'], 'percentRestriction': restrict['percentRestriction']}
			# data histogram
				tmpFeatureList = []
				for h in ag.sourceHandles:
					tmpFeatureList.append(h.data['raw'][restrict['anal']][0])
				tmpSlicedList = sliceListIntoEvenlySpacedLists(tmpFeatureList, (maxVal-minVal)/200.0, minVal, maxVal+10)
				distribution = []
				for i in tmpSlicedList: distribution.append(len(i))
			# printing
			for i, v in self.sourceTimeRestrictAnal.iteritems():
				print i, '->', v
	############################################################################
	def updateRestrict(self, chosenSampleHandle, chosenSampleData):
		# voice per second restriction...
		if not self.timesVoiceWasPlayed.has_key(chosenSampleHandle.voiceID):
			self.timesVoiceWasPlayed[chosenSampleHandle.voiceID] = []
		self.timesVoiceWasPlayed[chosenSampleHandle.voiceID].append(tgt.timeInFrames)
		# time varying restriction...
		if len(ops.CORPUS_RESTRICT_DISCRIPTOR_IN_TIME) > 0:
			pickedStream = chosenSampleData['availableStreams'][0]
			for anal in self.sourceTimeRestrictAnal: # record for each user-supplied feature...
				self.sourceTimeRestrictAnal[anal][pickedStream]['lastMetircVal'] = chosenSampleHandle.data['raw'][anal][0]
				self.sourceTimeRestrictAnal[anal][pickedStream]['lastActive'] = tgt.timeInFrames
	############################################################################
	def timeRestrictTest(self, sampleHandle):
		if not self.restrictTimeVarying:
			return [0] # returning stream 0, so everything is enabled
		# if passing this point, perform restriction test...
		availableStreams = []
		for anal in self.sourceTimeRestrictAnal: # test for each user-supplied feature...
			for streamID in self.sourceTimeRestrictAnal[anal]: # test for each stream...
				if self.sourceTimeRestrictAnal[anal][streamID]['lastMetircVal'] == None: 
					availableStreams.append(streamID) # don't restrict if not set yet by updateRestrict()
				else: # then test restriction...
					timeDifference = ag.f2s((tgt.timeInFrames-self.sourceTimeRestrictAnal[anal][streamID]['lastActive']))
					timeRatioToOnsets = timeDifference/self.sourceTimeRestrictAnal[anal][streamID]['restrictionReleaseSec']
					percentPossibleDueToTime = interpolateVal(timeRatioToOnsets, self.sourceTimeRestrictAnal[anal][streamID]['percentRestriction'], 100)
					if percentPossibleDueToTime >= 100: # then its over 100%, definitely do it
						availableStreams.append(streamID) 
						continue
					minVal = float(self.sourceTimeRestrictAnal[anal][streamID]['minVal'])
					maxVal = float(self.sourceTimeRestrictAnal[anal][streamID]['maxVal'])
					lastInThisStream = self.sourceTimeRestrictAnal[anal][streamID]['lastMetircVal']
					thisSample = sampleHandle.data['raw'][anal][0]
					absRawDiff = abs(lastInThisStream-thisSample) # difference of two metrics as a raw value
					absRawDiffAsPercentOfRange = (absRawDiff/(maxVal-minVal))*100.0
					if absRawDiffAsPercentOfRange <= percentPossibleDueToTime: # then its within the window, so add it
					#	print "PASSED", sampleHandle.data['raw'][anal][0], absRawDiffAsPercentOfRange, percentPossibleDueToTime
						availableStreams.append(streamID)
					#else: 
					#	print "FFFFFF", sampleHandle.data['raw'][anal][0], absRawDiffAsPercentOfRange, percentPossibleDueToTime
		#	print len(self.sourceTimeRestrictAnal)
		return availableStreams
	############################################################################
#	def printCorpusInfo(self):
#		postLevel = 2
#		printNumbSegsOrFiles = True
#		printLimits = True
#		# get number of total potential segments
#		totalSegs = 0
#		for name, dict in self.corpusPrintData.iteritems(): totalSegs += int(dict['segs'])
#		p.middle('Corpus - '+str(totalSegs)+' Possible Segments', postLevel)
#		
#		for name, dict in self.corpusPrintData.iteritems():
#			if dict['segFile'] != None: # make a printable list of segmentation files
#				tmp = []
#				for item in dict['segFile']:
#					tmp.append(os.path.split(item)[1])
#				name = ','.join(tmp)
#			
#			end = dict['limitEnd']
#			if end == sys.maxint: end = 'end'
#			printStr = "${YELLOW}"+str(dict['fileType'])+"${NORMAL} "+str(name)+" -> "
#			if printNumbSegsOrFiles:
#				if dict['fileType'] == 'file':
#					printStr += "${RED}"+dict['segs']+"${NORMAL} segments"
#					if dict['limitStart'] != 0 and str(end) not in ["end", None, "none"]:
#						printStr += " from "+str(dict['limitStart'])+" to "+str(end)+"."
#				elif dict['fileType'] == 'dir': printStr += "${RED}"+dict['segs']+"${NORMAL} segments from "+str(dict['numbSfFiles'])+" files."
#				elif dict['fileType'] == 'name': printStr += "${RED}"+dict['segs']+"${NORMAL} segments"
#				
#			if printLimits:
#				if len(dict['limit']) > 0:
#					for key, valDict in dict['limit'].iteritems():
#						restrict = [None, None]
#						for item, value in valDict.iteritems():
#							if item == 'scope': scoped = value
#							if item == 'low': restrict[0] = str(value)
#							if item == 'high': restrict[1] = str(value)
#							if item == 'lowPer': restrict[0] = str(value)+"%"
#							if item == 'highPer': restrict[1] = str(value)+"%"
#						if restrict[1] == None: restStr = "above "+str(restrict[0])
#						elif restrict[0] == None: restStr = "below "+restrict[1]
#						else: restStr = "between "+restrict[0]+" -> "+restrict[1]
#						printStr += "\n\t${RED}"+key+ "${NORMAL} "+restStr+", scoped ${YELLOW}"+scoped+"ly${NORMAL}"
#				if dict.has_key('restrictInTime') and dict['restrictInTime'] > 0:
#					print "\tRestricting segments to once every", dict['restrictInTime'], "seconds."
#				if dict['restrictOverlaps'] != None:
#					printStr += "\n\tRestricting segments to a ${RED}maximum simultanous overlap${NORMAL} of "+str(dict['restrictOverlaps'])+" -- scoped ${YELLOW}"+dict['restrictOverlapsReason']+"ly${NORMAL}."
#
#			p.post(printStr, postLevel)
#		p.post( "" )
#		pass
	############################################################################
	def printSampleHistogram(self):
		postLevel = 2
		noteHist = utilities.histogram(self.nameSelectedList)
		if len(noteHist) <= 1: noteHist = utilities.histogram(self.fileSelectedList)
		if ag.countRanOutOfSamples > 0: noteHist.append((ag.countRanOutOfSamples, "Out of Valid Segments"))
		p.middle('File Selection Histogram', postLevel)
		for (val, key) in noteHist:
			p.post( key+" -> ${RED}"+str(utilities.trunc((val/float(outputFiles.totalOutputNotes+ag.countRanOutOfSamples))*100, 2))+"%${NORMAL}", postLevel)
		p.post("", postLevel)
	############################################################################
	def printSimultanousHistogram(self):
		postLevel = 2
		tmpy = []
		for numbSeg in tgt.onsetsPerFrame:
			if numbSeg > 0: tmpy.append(numbSeg)
		#print tmpy
		noteHist = utilities.histogram(tmpy)
		#print noteHist
		p.middle('Simultaneous Selection Histogram', postLevel)
		for (val, key) in noteHist:
			p.post(str(key)+" note(s) -> ${RED}"+str(utilities.trunc(val/float(len(tmpy))*100, 2))+"%${NORMAL}", postLevel)
		p.post("", postLevel)
